#!/usr/bin/env python
# pylint: disable=E1103, E1101, C0103
# E1103: Use DB objects attached to thread
# E1101: Create config sections
# C0103: Internal methods start with _
"""
_BossAirAPI_

The interface for BossAir


Notes on convention:

BossAirAPI talks both inside and outside of the BossAir application family.
Interfaces geared toward the outside expect WMBS objects.  Interfaces
geared toward the inside expect RunJob objects.  Interior interfaces are
marked by names starting with '_' such as '_listRunning'
"""
import os.path
import threading
import logging
import subprocess
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.DAOFactory import DAOFactory
from WMCore.WMFactory import WMFactory
from WMCore.BossAir.RunJob import RunJob
from WMCore.WMConnectionBase import WMConnectionBase
from WMCore.WMException import WMException
from WMCore.FwkJobReport.Report import Report
from WMCore.WMExceptions import WM_JOB_ERROR_CODES


class BossAirException(WMException):
    """
    I expect you'll be seeing this a lot

    """


class BossAirAPI(WMConnectionBase):
    """
    _BossAirAPI_

    The API layer for the BossAir prototype
    """

    def __init__(self, config, insertStates=False):
        """
        __init__

        BossAir should work with the standard config
        structure of WMAgent
        """

        WMConnectionBase.__init__(self, daoPackage="WMCore.BossAir")

        myThread = threading.currentThread()

        self.config = config
        self.plugins = {}
        self.states = []

        self.jobs = []

        self.pluginDir = config.BossAir.pluginDir
        # This is the default state jobs are created in
        self.newState = getattr(config.BossAir, 'newState', 'New')

        # Get any proxy info
        self.checkProxy = getattr(config.BossAir, 'checkProxy', False)
        self.cert = getattr(config.BossAir, 'cert', None)

        self.stateMachine = ChangeState(self.config)

        # Create a factory to load plugins
        self.pluginFactory = WMFactory("plugins", self.pluginDir)

        self.daoFactory = DAOFactory(package="WMCore.BossAir",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        self.deleteDAO = self.daoFactory(classname="DeleteJobs")
        self.stateDAO = self.daoFactory(classname="NewState")
        self.loadByWMBSDAO = self.daoFactory(classname="LoadByWMBSID")
        self.updateDAO = self.daoFactory(classname="UpdateJobs")
        self.newJobDAO = self.daoFactory(classname="NewJobs")
        self.runningJobDAO = self.daoFactory(classname="LoadRunning")
        self.completeJobDAO = self.daoFactory(classname="LoadComplete")
        self.loadJobsDAO = self.daoFactory(classname="LoadByStatus")
        self.completeDAO = self.daoFactory(classname="CompleteJob")
        self.monitorDAO = self.daoFactory(classname="JobStatusForMonitoring")

        self.states = None
        self.loadPlugin(insertStates)

        return

    def loadPlugin(self, insertStates):
        """
        _loadPlugin_

        Actually load the plugin and init the database
        """

        states = set()

        for name in self.config.BossAir.pluginNames:
            self.plugins[name] = self.pluginFactory.loadObject(classname=name,
                                                               args=self.config)
            for state in self.plugins[name].states:
                states.add(state)

        if self.newState not in states:
            states.add(self.newState)

        if insertStates:
            # Add states only if we're not
            # doing a secondary instantiation
            self.addStates(states=states)

        self.states = states

        return

    def addStates(self, states):
        """
        _addStates_

        Add States to bl_status table. Meant to be done only
        once in an agent lifetime.
        """
        existingTransaction = self.beginTransaction()

        self.stateDAO.execute(states=states, conn=self.getDBConn(),
                              transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)

        return

    def createNewJobs(self, wmbsJobs):
        """
        _createNewJobs_

        Create new jobs in the BossAir database
        Accepts WMBS Jobs
        """

        existingTransaction = self.beginTransaction()

        jobsToCreate = []

        # First turn wmbsJobs into runJobs
        for wmbsJob in wmbsJobs:
            runJob = RunJob()
            runJob.buildFromJob(job=wmbsJob)
            if runJob.get('status') not in self.states:
                runJob['status'] = self.newState
            jobsToCreate.append(runJob)

        # Next insert them into the database
        self.newJobDAO.execute(jobs=jobsToCreate, conn=self.getDBConn(),
                               transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)

        return

    def _listRunJobs(self, active=True):
        """
        _listRunJobs_

        List runjobs, either active or complete
        """

        if active:
            runJobDicts = self.runningJobDAO.execute(conn=self.getDBConn(),
                                                     transaction=self.existingTransaction())
        else:
            runJobDicts = self.completeJobDAO.execute(conn=self.getDBConn(),
                                                      transaction=self.existingTransaction())

        runJobs = []
        for jDict in runJobDicts:
            rj = RunJob()
            rj.update(jDict)
            runJobs.append(rj)

        return runJobs

    def _loadByStatus(self, status, complete='1'):
        """
        _loadByStatus_

        Load jobs by status
        """

        if status not in self.states:
            msg = "Asked to load by status %s which is not loaded\n" % (status)
            msg += "This indicates that the wrong plugins are loaded\n"
            logging.error(msg)
            raise BossAirException(msg)

        loadJobs = self.loadJobsDAO.execute(status=status,
                                            complete=complete,
                                            conn=self.getDBConn(),
                                            transaction=self.existingTransaction())
        statusJobs = []
        for jDict in loadJobs:
            rj = RunJob()
            rj.update(jDict)
            statusJobs.append(rj)

        return statusJobs

    def _loadByID(self, jobs):
        """
        _loadByID_

        Load by running Job ID
        """
        loadJobsDAO = self.daoFactory(classname="LoadByID")
        loadJobs = loadJobsDAO.execute(jobs=jobs, conn=self.getDBConn(),
                                       transaction=self.existingTransaction())

        loadedJobs = []
        for jDict in loadJobs:
            rj = RunJob()
            rj.update(jDict)
            loadedJobs.append(rj)

        return loadedJobs

    def _updateJobs(self, jobs):
        """
        _updateJobs_

        Update the job entries in the BossAir database
        """

        if len(jobs) < 1:
            # Nothing to do
            return

        existingTransaction = self.beginTransaction()

        self.updateDAO.execute(jobs=jobs, conn=self.getDBConn(),
                               transaction=self.existingTransaction())

        jobsWithLocation = [job for job in jobs if job.get('location') is not None]
        if jobsWithLocation:
            self.stateMachine.recordLocationChange(jobsWithLocation)

        self.commitTransaction(existingTransaction)

        return

    def _deleteJobs(self, jobs):
        """
        _deleteJobs_

        Delete the job entries in the BossAir database
        NOTE: only used by unit tests
        """

        if len(jobs) < 1:
            # Nothing to do
            return

        idList = [x['id'] for x in jobs]

        existingTransaction = self.beginTransaction()

        self.deleteDAO.execute(jobs=idList, conn=self.getDBConn(),
                               transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)

        return

    def loadByWMBS(self, wmbsJobs):
        """
        _loadByWMBS_

        Load BossAir info based on wmbs Jobs.
        """

        if len(wmbsJobs) < 1:
            return []

        jobList = self.loadByWMBSDAO.execute(jobs=wmbsJobs, conn=self.getDBConn(),
                                             transaction=self.existingTransaction())

        loadedJobs = []
        for job in jobList:
            rj = RunJob()
            rj.buildFromJob(job)
            loadedJobs.append(rj)

        if len(loadedJobs) != len(wmbsJobs):
            logging.error("Could not load all jobs in BossAir for WMBS input!")
            idList = [x['jobid'] for x in loadedJobs]
            for job in wmbsJobs:
                if job['id'] not in idList:
                    logging.error("Failed to retrieve wmbs_id %i and WMBS job info: %s", job['id'], job)

        return loadedJobs

    def check(self):
        """
        _check_

        Perform checks of critical components, i.e. proxy validation, etc.
        """

        if self.checkProxy:
            command = 'voms-proxy-info'
            if self.cert is not None and self.cert != '':
                command += ' --file ' + self.cert

            pipe = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, dummyerr = pipe.communicate()

            try:
                output = output.split("timeleft  :")[1].strip()
            except IndexError:
                raise BossAirException("Missing Proxy", output.strip())

            if output == "0:00:00":
                raise BossAirException("Proxy Expired", output.strip())

        return

    def submit(self, jobs, info=None):
        """
        _submit_

        Submit jobs using the plugin

        Requires both plugin name and workflow user from submitter

        Deals internally in RunJob objects, but interfaces
        to the outside with WMBS Job analogs

        Returns (successes, failures)
        """
        self.check()

        successJobs = []
        failureJobs = []

        # TODO: Add plugin and user to input via JobSubmitter
        # IMPORTANT IMPORTANT IMPORTANT

        # Put job into RunJob format
        pluginDict = {}

        for job in jobs:
            rj = RunJob()
            rj.buildFromJob(job=job)
            if not job.get('location', False):
                rj['location'] = job.get('custom', {}).get('location', None)
            plugin = rj['plugin']
            pluginDict.setdefault(plugin, [])
            pluginDict[plugin].append(rj)
            # Can't add to the cache in submit()
            # It's NOT the same bossAir instance
            # self.jobs.append(rj)

        for plugin in list(pluginDict):  # .clear() makes pluginDict change size during iteration!
            if plugin not in self.plugins:
                # Then we have a non-existant plugin
                msg = "CRITICAL ERROR: Non-existant plugin!\n"
                msg += "Given a plugin %s that we don't have access to.\n" % (plugin)
                msg += "Ignoring the jobs for this plugin for now"
                logging.error(msg)
                continue
            try:
                pluginInst = self.plugins[plugin]
                jobsToSubmit = pluginDict.get(plugin, [])
                logging.debug("About to submit %i jobs to plugin %s", len(jobsToSubmit), plugin)
                localSuccess, localFailure = pluginInst.submit(jobs=jobsToSubmit,
                                                               info=info)
                for job in localSuccess:
                    successJobs.append(job.buildWMBSJob())
                for job in localFailure:
                    failureJobs.append(job.buildWMBSJob())
            except WMException:
                raise
            except Exception as ex:
                msg = "Unhandled exception while submitting jobs to plugin: %s\n" % plugin
                msg += str(ex)
                logging.error(msg)
                logging.debug("Jobs being submitted: %s\n", jobsToSubmit)
                logging.debug("Job info: %s\n", info)
                raise BossAirException(msg)
            finally:
                # make sure we release this memory
                pluginDict.clear()
                del jobsToSubmit[:]

        # Create successful jobs in BossAir
        try:
            logging.debug("About to create %i new jobs in BossAir", len(successJobs))
            self.createNewJobs(wmbsJobs=successJobs)
        except WMException:
            raise
        except Exception as ex:
            msg = "Unhandled error in creation of %i new jobs.\n" % len(successJobs)
            msg += str(ex)
            logging.error(msg)
            logging.debug("Job: %s", successJobs)
            raise BossAirException(msg)

        return successJobs, failureJobs

    def track(self, runJobIDs=None, wmbsIDs=None):
        """
        _track_

        Track all running jobs
        Load job info from the cache (it should be there since we submitted the job)

        OPTIONAL: You can submit a list of jobs to check, based either on wmbsIDs or
         on runjobIDs.  This takes a list of integer IDs.
        """

        jobsToChange = []
        jobsToComplete = []
        jobsToReturn = []
        returnList = []

        jobsToTrack = {}

        runningJobs = self._listRunJobs(active=True)

        if runJobIDs:
            for job in runningJobs:
                if job['id'] not in runJobIDs:
                    runningJobs.remove(job)
        if wmbsIDs:
            for job in runningJobs:
                if job['jobid'] not in wmbsIDs:
                    runningJobs.remove(job)

        if len(runningJobs) < 1:
            # Then we have no running jobs
            return returnList

        logging.info("About to start building running jobs")

        loadedJobs = self._buildRunningJobsFromRunJobs(runJobs=runningJobs)

        logging.info("About to look for %i loadedJobs.", len(loadedJobs))

        for runningJob in loadedJobs:
            plugin = runningJob['plugin']
            if plugin not in jobsToTrack:
                jobsToTrack[plugin] = []
            jobsToTrack[plugin].append(runningJob)

        for plugin in jobsToTrack:
            if plugin not in self.plugins:
                msg = "Jobs tracking with non-existant plugin %s\n" % (plugin)
                msg += "They were submitted but can't be tracked?\n"
                msg += "That's too strange to continue\n"
                logging.error(msg)
                raise BossAirException(msg)
            try:
                # Then we send them to the plugins
                # Should give you a lit of jobs to change and jobs to complete
                pluginInst = self.plugins[plugin]
                localRunning, localChanges, localCompletes = pluginInst.track(jobs=jobsToTrack[plugin])
                jobsToReturn.extend(localRunning)
                jobsToChange.extend(localChanges)
                jobsToComplete.extend(localCompletes)
                logging.info("Executing/changing/completing %i/%i/%i jobs in plugin %s.", len(localRunning),
                             len(localChanges), len(localCompletes), plugin)
            except WMException:
                raise
            except Exception as ex:
                msg = "Unhandled exception while tracking jobs for plugin %s!\n" % plugin
                msg += str(ex)
                logging.error(msg)
                logging.debug("JobsToTrack: %s", jobsToTrack[plugin])
                raise BossAirException(msg)

        logging.info("About to change %i jobs", len(jobsToChange))
        logging.debug("JobsToChange: %s", jobsToChange)
        logging.info("About to complete %i jobs", len(jobsToComplete))
        logging.debug("JobsToComplete: %s", jobsToComplete)

        self._updateJobs(jobs=jobsToChange)
        self._complete(jobs=jobsToComplete)

        # We should have a globalState variable for changed jobs
        # from the plugin
        # Return that to the calling function
        for rj in jobsToReturn:
            job = rj.buildWMBSJob()
            job['globalState'] = rj['globalState']
            returnList.append(job)

        return returnList

    def _complete(self, jobs):
        """
        _complete_

        Complete jobs using plugin functions
        Requires jobs in RunJob format
        """
        if len(jobs) < 1:
            return

        # We should be insulated from bad plugins by track()
        jobsToComplete = {}

        for job in jobs:
            if job['plugin'] not in jobsToComplete:
                jobsToComplete[job['plugin']] = []
            jobsToComplete[job['plugin']].append(job)

        try:
            for plugin in jobsToComplete:
                self.plugins[plugin].complete(jobsToComplete[plugin])
        except WMException:
            raise
        except Exception as ex:
            msg = "Exception while completing jobs!\n"
            msg += str(ex)
            logging.error(msg)
            logging.debug("JobsToComplete: %s", jobsToComplete)
            raise BossAirException(msg)
        finally:
            # If the complete code fails, label the jobs as finished anyway
            # We want to avoid cyclic repetition of failed jobs
            # If they don't have a FWJR, the Accountant will catch it.
            self._completeKill(jobs)

        return

    def _completeKill(self, jobs):
        """
        __completeKill_

        Mark jobs killed in BossAir as completed
        Requires jobs in RunJob format
        """
        if len(jobs) < 1:
            return

        idsToComplete = [job['id'] for job in jobs]

        existingTransaction = self.beginTransaction()
        self.completeDAO.execute(jobs=idsToComplete, conn=self.getDBConn(),
                                 transaction=self.existingTransaction())
        self.commitTransaction(existingTransaction)

        return

    def getComplete(self):
        """
        _getComplete_

        The tracker should call this: It's only
        interested in the jobs that are completed.
        """

        completeJobs = []

        completeRunJobs = self._listRunJobs(active=False)

        for rj in completeRunJobs:
            job = rj.buildWMBSJob()
            completeJobs.append(job)

        return completeJobs

    def kill(self, jobs, workflowName=None, killMsg=None, errorCode=71300):
        """
        _kill_

        Kill jobs using plugin functions:

        Only active jobs (status = 1) will be killed. If workflowName is given,
        then kill all its jobs in one shot.
        An optional killMsg can be sent; this will be written into the job FWJR.
        The errorCode will be the one specified and if no killMsg is provided then
        a standard message associated with the exit code will be used.
        If a previous FWJR exists, this error will be appended to it.
        """
        if not jobs:
            return
        jobsToKill = {}

        # Now get a list of which jobs are in the batch system
        # only kill jobs present there
        loadedJobs = self._buildRunningJobs(wmbsJobs=jobs)

        for runningJob in loadedJobs:
            plugin = runningJob['plugin']
            jobsToKill.setdefault(plugin, [])
            jobsToKill[plugin].append(runningJob)

        for plugin in jobsToKill:
            if plugin not in self.plugins:
                msg = "Jobs tracking with non-existant plugin %s\n" % (plugin)
                msg += "They were submitted but can't be tracked?\n"
                msg += "That's too strange to continue\n"
                logging.error(msg)
                raise BossAirException(msg)
            else:
                # Then we send them to the plugins
                try:
                    pluginInst = self.plugins[plugin]
                    if workflowName:
                        # jobs are completed regardless whether the kill succeeded or not
                        self._completeKill(jobs=jobsToKill[plugin])
                        pluginInst.killWorkflowJobs(workflow=workflowName)
                    else:
                        # raise an exception if it fails to kill jobs, such that the same
                        # jobs are retried again in the next cycle
                        pluginInst.kill(jobs=jobsToKill[plugin], raiseEx=True)
                        self._completeKill(jobs=jobsToKill[plugin])

                    # Register the killed jobs
                    for job in jobsToKill[plugin]:
                        if job.get('cache_dir') is None or job.get('retry_count') is None:
                            continue
                        # Try to save an error report as the jobFWJR
                        if not os.path.isdir(job['cache_dir']):
                            # Then we have a bad cache directory
                            logging.error("Could not write a kill FWJR due to non-existant cache_dir for job %i\n", job['id'])
                            logging.debug("cache_dir: %s\n", job['cache_dir'])
                            continue
                        reportName = os.path.join(job['cache_dir'], 'Report.%i.pkl' % job['retry_count'])
                        errorReport = Report()
                        if os.path.exists(reportName) and os.path.getsize(reportName) > 0:
                            # Then there's already a report there.  Add messages
                            errorReport.load(reportName)
                        # Build a better job message
                        if killMsg:
                            reportedMsg = killMsg
                        else:
                            reportedMsg = WM_JOB_ERROR_CODES[errorCode]
                            reportedMsg += '\n Job last known status was: %s' % job.get('globalState', 'Unknown')
                        errorReport.addError("JobKilled", errorCode, "JobKilled", reportedMsg)
                        try:
                            errorReport.save(filename=reportName)
                        except IOError as ioe:
                            logging.warning('Cannot write report %s because of %s', reportName, ioe)
                except RuntimeError:
                    logging.warning("Plugin failed to remove jobs. It will be retried in the next cycle.")
                except WMException:
                    raise
                except Exception as ex:
                    msg = "Unhandled exception while calling kill method for plugin %s\n" % plugin
                    msg += str(ex)
                    logging.error(msg)
                    logging.debug("Interrupted while killing following jobs: %s\n", jobsToKill[plugin])
                    raise BossAirException(msg)
        return

    def update(self, jobs):
        """
        _update_

        Overwrite the database with whatever you put into
        this function.
        """

        runJobs = self._buildRunningJobs(wmbsJobs=jobs)

        self._updateJobs(jobs=runJobs)

        return

    def monitor(self, commonState=True):
        """
        _monitor_

        Initiate the call to the monitoring DAO
        This should not be called by the standard Submitter/Status/Tracker
        system.  It is meant for outside calling.
        """
        results = self.monitorDAO.execute(commonState, conn=self.getDBConn(),
                                          transaction=self.existingTransaction())

        return results

    def updateJobInformation(self, workflow, **kwargs):
        """
        _updateJobInformation_

        Update the information of jobs in a particular workflow,
        the data will be updated according the keyword arguments which
        will be interpreted by the individual plugins accordingly.
        """
        for plugin in self.plugins:
            try:
                pluginInst = self.plugins[plugin]
                pluginInst.updateJobInformation(workflow, **kwargs)
            except WMException:
                raise
            except Exception as ex:
                msg = "Unhandled exception while calling update method for plugin %s\n" % plugin
                msg += str(ex)
                logging.error(msg)
                raise BossAirException(msg)
        return

    def updateSiteInformation(self, jobs, siteName, excludeSite):
        """
        _updateSiteInformation_

        Modify condor classAd for all Idle jobs for a site if it has gone Down, Draining or Aborted.
        Kill all jobs if the site is the only site for the job.
        """
        jobkill = []
        for plugin in self.plugins:
            try:
                pluginInst = self.plugins[plugin]
                tempjoblist = pluginInst.updateSiteInformation(jobs, siteName, excludeSite)
                if tempjoblist is not None:
                    jobkill.extend(tempjoblist)
            except WMException:
                raise
            except Exception as ex:
                msg = "Unhandled exception while calling update method for plugin %s\n" % plugin
                msg += str(ex)
                logging.error(msg)
                raise BossAirException(msg)
        return jobkill

    def _buildRunningJobsFromRunJobs(self, runJobs):
        """
        _buildRunningJobsFromRunJobs_

        Same as _buildRunningJobs_, but taking runJobs as input
        """
        finalJobs = []

        loadedJobs = self._loadByID(jobs=runJobs)

        for loadJob in loadedJobs:
            runJob = None
            for rj in runJobs:
                if rj['id'] == loadJob['id']:
                    runJob = rj
                    break
            # We should have two instances of the job
            for key in runJob:
                # Fill one from the other
                # runJob, being most recent, should be on top
                if runJob[key] is None:
                    runJob[key] = loadJob.get(key, None)
            finalJobs.append(runJob)

        return finalJobs

    def _buildRunningJobs(self, wmbsJobs):
        """
        _buildRunningJobs_

        Build running jobs by loading information from the database and
        compiling it into a runJob object.  This overwrites any information
        from the database with the info from the WMBS Job
        """
        runJobsLoaded = self.loadByWMBS(wmbsJobs=wmbsJobs)

        return runJobsLoaded



#!/usr/bin/env python
"""
_CountUndeletedBlocksByWorkflow_

Oracle implementation of Workflow.CountUndeletedBlocksByWorkflow
"""

from WMComponent.DBS3Buffer.MySQL.CountUndeletedBlocksByWorkflow import CountUndeletedBlocksByWorkflow as MySQLCountUndeletedBlocksByWorkflow


class CountUndeletedBlocksByWorkflow(MySQLCountUndeletedBlocksByWorkflow):
    """
    Retrieves a list of all workflows and the relative deleted blocks counters
    """



#!/usr/bin/env python
"""
_FailedJobs_

Oracle implementation of Workflow.FailedJobs
"""




from WMCore.WMBS.MySQL.Workflow.FailedJobs import FailedJobs as MySQLFailedJobs

class FailedJobs(MySQLFailedJobs):
    pass



#!/usr/bin/env python
"""
_GetLocation_

MySQL implementation of Jobs.GetAllJobs
"""

from WMCore.Database.DBFormatter import DBFormatter

from future.utils import listvalues

class GetAllJobs(DBFormatter):
    """
    _GetLocation_

    Retrieve all files that are associated with the given job from the
    database.
    """
    sql_all = "SELECT id FROM wmbs_job"

    sql_state = "SELECT id FROM wmbs_job WHERE state = (SELECT id FROM wmbs_job_state WHERE name = :state)"

    sql_state_type = """SELECT wmbs_job.id FROM wmbs_job
                          INNER JOIN wmbs_jobgroup ON wmbs_job.jobgroup = wmbs_jobgroup.id
                          INNER JOIN wmbs_subscription ON wmbs_jobgroup.subscription = wmbs_subscription.id
                          INNER JOIN wmbs_job_state ON wmbs_job.state = wmbs_job_state.id
                          INNER JOIN wmbs_sub_types ON wmbs_subscription.subtype = wmbs_sub_types.id
                          WHERE wmbs_job_state.name = :state
                          AND wmbs_sub_types.name = :type"""

    limit_sql = " limit %d"

    def format(self, results):
        """
        _formatDict_

        Cast the file attribute to an integer, and also handle changing the
        column name in Oracle from FILEID to FILE.
        """

        if len(results) == 0:
            return False
        else:
            tempList = results[0].fetchall()
            final = []
            for i in tempList:
                final.append(listvalues(i)[0])
            return final

    def execute(self, state=None, jobType=None, conn=None,
                transaction=False, limitRows=None):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """
        if limitRows:
            extraSql = self.limit_sql % limitRows
        else:
            extraSql = ""

        if state is None:
            result = self.dbi.processData(self.sql_all + extraSql, {}, conn=conn,
                                          transaction=transaction)
        else:
            if jobType:
                result = self.dbi.processData(self.sql_state_type + extraSql, {'state': state.lower(), 'type': jobType},
                                              conn=conn, transaction=transaction)
            else:
                result = self.dbi.processData(self.sql_state + extraSql, {'state': state.lower()},
                                              conn=conn, transaction=transaction)

        res = self.format(result)
        return res



"""
_PeriodicSiblingComplete_

Oracle implementation of JobSplitting.PeriodicSiblingComplete
"""

from WMCore.WMBS.MySQL.JobSplitting.PeriodicSiblingComplete import PeriodicSiblingComplete as MySQLPeriodicSiblingComplete

class PeriodicSiblingComplete(MySQLPeriodicSiblingComplete):
    """
    Right now this is the same as the MySQL version.

    """



#!/usr/bin/env python
"""
_GetFinishedTasks_

Oracle implementation of Workflow.GetFinishedTasks

Created on Nov 7, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Workflow.GetFinishedTasks import GetFinishedTasks as MySQLGetFinishedTasks

class GetFinishedTasks(MySQLGetFinishedTasks):
    pass



#!/usr/bin/env python
"""
_StageOut_

Diagnostic implementation for a job StageOut


"""

from builtins import range
import os
from WMCore.WMSpec.Steps.Diagnostic import Diagnostic, DiagnosticHandler


class SOMExceptionHandler(DiagnosticHandler):
    """
    _SOMExceptionHandler_

    Generic handler for the StageOutManager

    I have no idea what this should do

    """
    def __call__(self, errCode, executor, **args):
        """
        _operator()_

        Twiddle thumbs, contemplate navel, toss coin

        """
        msg         = "Error in StageOut: %s\n" % (errCode)
        description = "Misc. StageOut error: %s\n" % (errCode)
        if args.get('ex', None):
            msg += str(args.get('ex'))
        if args.get('ExceptionInstance', False):
            msg += str(args.get('ExceptionInstance'))

        #jobRepXml = os.path.join(executor.step.builder.workingDir,
        #                         executor.step.output.jobReport)

        #if not os.path.exists(jobRepXml):
        #    # no report => Error
        #    msg = "No Job Report Found: %s" % jobRepXml
        #    executor.report.addError(50115, "MissingJobReport", msg)
        #    return

        # job report XML exists, load the exception information from it
        #executor.report.parse(jobRepXml)


        # make sure the report has the error in it
        errSection = getattr(executor.report.report, "errors", None)
        if errSection == None:
            localmsg = "Job Report contains no error report, but StageOutManager exited non-zero: %s" % errCode
            executor.report.addError(executor.stepName, 50116, "MissingErrorReport", localmsg)

        else:
            #check exit code in report is non zero
            if executor.report.report.status == 0:
                msg = "Job Report contains no error report, but StageOutManager exited non-zero: %s" % errCode


        executor.report.addError(executor.stepName,
                                 errCode, description, msg)

        return

class StageOut(Diagnostic):

    def __init__(self):
        Diagnostic.__init__(self)

        # Set defaults
        catchAll            = SOMExceptionHandler()
        self.defaultHandler = catchAll

        [ self.handlers.__setitem__(x, catchAll) for x in range(0, 255) if x not in self.handlers ]



#!/usr/bin/env python
#pylint: disable=E1103
# E1103:  Attach objects to threading
"""
_TwoFileBased_

File based job splitting for two file read workflows.  This works the same as
normal file based splitting except that the input files will also have their
parentage information loaded so that the parents can be included in the job.
"""




import logging
import threading

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.DAOFactory              import DAOFactory

from WMCore.WMBS.File  import File

class TwoFileBased(JobFactory):
    """
    Two file read workflow splitting

    """


    def __init__(self, package='WMCore.DataStructs',
                 subscription=None,
                 generators=[],
                 limit = None):
        """
        __init__

        Create the DAOs
        """

        myThread = threading.currentThread()

        JobFactory.__init__(self, package = 'WMCore.WMBS',
                            subscription = subscription,
                            generators = generators,
                            limit = limit)


        self.daoFactory = DAOFactory(package = "WMCore.WMBS",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)

        self.getParentInfoAction     = self.daoFactory(classname = "Files.GetParentAndGrandParentInfo")



        return



    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Split up all the available files such that each job will process a
        maximum of "files_per_job".  If the "files_per_job" parameters is not
        passed in jobs will process a maximum of 10 files.
        """
        filesPerJob = int(kwargs.get("files_per_job", 10))
        jobsPerGroup = int(kwargs.get("jobs_per_group", 0))
        filesInJob   = 0
        totalJobs    = 0
        listOfFiles  = []

        #Get a dictionary of sites, files
        locationDict = self.sortByLocation()

        for location in locationDict:
            #Now we have all the files in a certain location
            fileList    = locationDict[location]
            filesInJob  = 0
            jobsInGroup = 0
            self.newGroup()
            if len(fileList) == 0:
                #No files for this location
                #This isn't supposed to happen, but better safe then sorry
                logging.debug("Have location %s with no files" % (location))
                continue
            for file in fileList:
                parentLFNs = self.findParent(lfn = file['lfn'])
                for lfn in parentLFNs:
                    parent = File(lfn = lfn)
                    file['parents'].add(parent)
                if filesInJob == 0 or filesInJob == filesPerJob:
                    if jobsPerGroup:
                        if jobsInGroup > jobsPerGroup:
                            self.newGroup()
                            jobsInGroup = 0

                    self.newJob(name = self.getJobName(length=totalJobs))

                    filesInJob   = 0
                    totalJobs   += 1
                    jobsInGroup += 1

                filesInJob += 1
                self.currentJob.addFile(file)

                listOfFiles.append(file)




        return




    def findParent(self, lfn):
        """
        _findParent_

        Find the parents for a file based on its lfn
        """


        parentsInfo = self.getParentInfoAction.execute([lfn])
        newParents = set()
        for parentInfo in parentsInfo:

            # This will catch straight to merge files that do not have redneck
            # parents.  We will mark the straight to merge file from the job
            # as a child of the merged parent.
            if int(parentInfo["merged"]) == 1:
                newParents.add(parentInfo["lfn"])

            elif parentInfo['gpmerged'] == None:
                continue

            # Handle the files that result from merge jobs that aren't redneck
            # children.  We have to setup parentage and then check on whether or
            # not this file has any redneck children and update their parentage
            # information.
            elif int(parentInfo["gpmerged"]) == 1:
                newParents.add(parentInfo["gplfn"])

            # If that didn't work, we've reached the great-grandparents
            # And we have to work via recursion
            else:
                parentSet = self.findParent(lfn = parentInfo['gplfn'])
                for parent in parentSet:
                    newParents.add(parent)

        return newParents



#!/usr/bin/env python
"""
_Step.Executor.StageOut_

Implementation of an Executor for a StageOut step

"""
from __future__ import print_function

import logging
import os
import os.path
import signal
import sys

from WMCore.Algorithms.Alarm import Alarm, alarmHandler
from WMCore.FwkJobReport.Report import Report
from WMCore.Lexicon import lfn     as lfnRegEx
from WMCore.Lexicon import userLfn as userLfnRegEx
from WMCore.Storage.FileManager import StageOutMgr as FMStageOutMgr
from WMCore.Storage.StageOutMgr import StageOutMgr
from WMCore.WMSpec.Steps.Executor import Executor


class StageOut(Executor):
    """
    _StageOut_

    Execute a StageOut Step

    """

    def pre(self, emulator=None):
        """
        _pre_

        Pre execution checks

        """

        # Are we using an emulator?
        if emulator is not None:
            return emulator.emulatePre(self.step)

        logging.info("Steps.Executors.%s.pre called", self.__class__.__name__)
        return None

    def execute(self, emulator=None):
        """
        _execute_


        """
        # Are we using emulators again?
        if emulator is not None:
            return emulator.emulate(self.step, self.job)

        logging.info("Steps.Executors.%s.execute called", self.__class__.__name__)

        overrides = {}
        if hasattr(self.step, 'override'):
            overrides = self.step.override.dictionary_()
        # propagete upstream cmsRun outcome such that we can decide whether to
        # stage files out or not
        self.failedPreviousStep = overrides.get('previousCmsRunFailure', False)

        # Set wait to two hours per retry
        # this alarm leaves a subprocess behing that may cause trouble, see #6273
        waitTime = overrides.get('waitTime', 7200 * self.step.retryCount)

        logging.info("StageOut override is: %s ", self.step)

        # Pull out StageOutMgr Overrides

        # switch between old stageOut behavior and new, fancy stage out behavior
        useNewStageOutCode = False
        if getattr(self.step, 'newStageout', False) or \
                ('newStageOut' in overrides and overrides.get('newStageOut')):
            useNewStageOutCode = True

        stageOutCall = {}
        if "command" in overrides and "option" in overrides \
                and "phedex-node" in overrides \
                and "lfn-prefix" in overrides:
            logging.critical('using override in StageOut')
            stageOutCall['command'] = overrides.get('command')
            stageOutCall['option'] = overrides.get('option')
            stageOutCall['phedex-node'] = overrides.get('phedex-node')
            stageOutCall['lfn-prefix'] = overrides.get('lfn-prefix')

        # naw man, this is real
        # iterate over all the incoming files
        if not useNewStageOutCode:
            # old style
            manager = StageOutMgr(**stageOutCall)
            manager.numberOfRetries = self.step.retryCount
            manager.retryPauseTime = self.step.retryDelay
        else:
            # new style
            logging.critical("STAGEOUT IS USING NEW STAGEOUT CODE")
            manager = FMStageOutMgr(retryPauseTime=self.step.retryDelay,
                                    numberOfRetries=self.step.retryCount,
                                    **stageOutCall)

        # We need to find a list of steps in our task
        # And eventually a list of jobReports for out steps

        # Search through steps for report files
        filesTransferred = []

        for step in self.stepSpace.taskSpace.stepSpaces():
            if step == self.stepName:
                # Don't try to parse your own report; it's not there yet
                continue
            stepLocation = os.path.join(self.stepSpace.taskSpace.location, step)
            logging.info("Beginning report processing for step %s", step)
            reportLocation = os.path.join(stepLocation, 'Report.pkl')
            if not os.path.isfile(reportLocation):
                logging.error("Cannot find report for step %s in space %s", step, stepLocation)
                continue
            # First, get everything from a file and 'unpersist' it
            stepReport = Report()
            stepReport.unpersist(reportLocation, step)

            # Don't stage out files from bad steps. Each step has its own Report.pkl file
            # We need to check all steps executed so far, otherwise it might stage out
            # files for chained steps when the overall job has already failed to process
            # one of them
            if not stepReport.stepSuccessful(step) or self.failedPreviousStep:
                msg = "Either the step did not succeed or an upstream step failed. "
                msg += "Skipping stage out of any root output files in this job."
                logging.warning(msg)
                continue

            # Okay, time to start using stuff
            # Now I'm a bit confused about this; each report should ONLY
            # Have the results of that particular step in it,
            # So getting all the files should get ONLY the files
            # for that step; or so I hope
            files = stepReport.getAllFileRefsFromStep(step=step)
            for fileName in files:

                # make sure the file information is consistent
                if hasattr(fileName, 'pfn') and (not hasattr(fileName, 'lfn') or not hasattr(fileName, 'module_label')):
                    msg = "Not a valid file: %s" % fileName
                    logging.error(msg)
                    continue

                # Figuring out if we should do straight to merge
                #  - should we do straight to merge at all ?
                #  - is straight to merge disabled for this output ?
                #  - are we over the size threshold
                #  - are we over the event threshold ?
                straightToMerge = False
                if not getattr(fileName, 'merged', False):
                    if hasattr(fileName, 'dataset') and fileName.dataset.get('dataTier', "") in ["NANOAOD",
                                                                                                 "NANOAODSIM"]:
                        logging.info("NANOAOD and NANOAODSIM files never go straight to merge!")
                    elif fileName.module_label not in getattr(self.step.output, 'forceUnmergedOutputs', []):
                        if hasattr(self.step.output, 'minMergeSize') and getattr(fileName, 'size',
                                                                                 0) >= self.step.output.minMergeSize:
                            logging.info("Sending %s straight to merge due to minMergeSize", fileName.lfn)
                            straightToMerge = True
                        elif getattr(fileName, 'events', 0) >= getattr(self.step.output, 'maxMergeEvents', sys.maxsize):
                            logging.info("Sending %s straight to merge due to maxMergeEvents", fileName.lfn)
                            straightToMerge = True

                if straightToMerge:
                    try:
                        fileName = self.handleLFNForMerge(mergefile=fileName, step=step)
                    except Exception as ex:
                        logging.info("minMergeSize: %s", getattr(self.step.output, 'minMergeSize', None))
                        logging.info("maxMergeEvents: %s", getattr(self.step.output, 'maxMergeEvents', None))
                        logging.error("Encountered error while handling LFN for merge %s", fileName)
                        logging.error(str(ex))
                        manager.cleanSuccessfulStageOuts()
                        stepReport.addError(self.stepName, 60401, "DirectToMergeFailure", str(ex))

                # Save the input PFN in case we need it
                # Undecided whether to move fileName.pfn to the output PFN
                fileName.InputPFN = fileName.pfn
                lfn = fileName.lfn
                fileSource = getattr(fileName, 'Source', None)
                if fileSource in ['TFileService', 'UserDefined']:
                    userLfnRegEx(lfn)
                else:
                    lfnRegEx(lfn)

                fileForTransfer = {'LFN': lfn,
                                   'PFN': getattr(fileName, 'pfn'),
                                   'PNN': None,
                                   'StageOutCommand': None,
                                   'Checksums': getattr(fileName, 'checksums', None)}

                signal.signal(signal.SIGALRM, alarmHandler)
                signal.alarm(waitTime)
                try:
                    manager(fileForTransfer)
                    # Afterwards, the file should have updated info.
                    filesTransferred.append(fileForTransfer)
                    fileName.StageOutCommand = fileForTransfer['StageOutCommand']
                    fileName.location = fileForTransfer['PNN']
                    fileName.OutputPFN = fileForTransfer['PFN']
                except Alarm:
                    msg = "Indefinite hang during stageOut of logArchive"
                    logging.error(msg)
                    manager.cleanSuccessfulStageOuts()
                    stepReport.addError(self.stepName, 60403, "StageOutTimeout", msg)
                    # well, if it fails for one file, it fails for the whole job...
                    break
                except Exception as ex:
                    manager.cleanSuccessfulStageOuts()
                    stepReport.addError(self.stepName, 60307, "StageOutFailure", str(ex))
                    stepReport.persist(reportLocation)
                    raise

                signal.alarm(0)

            # Am DONE with report. Persist it
            stepReport.persist(reportLocation)

        # Done with all steps, and should have a list of
        # stagedOut files in fileForTransfer
        logging.info("Transferred %i files", len(filesTransferred))
        return

    def post(self, emulator=None):
        """
        _post_

        Post execution checkpointing

        """
        # Another emulator check
        if emulator is not None:
            return emulator.emulatePost(self.step)

        logging.info("Steps.Executors.%s.post called", self.__class__.__name__)

        for step in self.stepSpace.taskSpace.stepSpaces():

            if step == self.stepName:
                # Don't try to parse your own report; it's not there yet
                continue

            stepLocation = os.path.join(self.stepSpace.taskSpace.location, step)
            logging.info("Beginning report processing for step %s", step)

            reportLocation = os.path.join(stepLocation, 'Report.pkl')
            if not os.path.isfile(reportLocation):
                logging.error("Cannot find report for step %s in space %s", step, stepLocation)
                continue

            # First, get everything from a file and 'unpersist' it
            stepReport = Report(step)
            stepReport.unpersist(reportLocation)

            # Don't stage out files from bad steps.
            if not stepReport.stepSuccessful(step):
                continue

            files = stepReport.getAllFileRefsFromStep(step=step)
            for fileInfo in files:
                if hasattr(fileInfo, 'lfn') and hasattr(fileInfo, 'location') and hasattr(fileInfo, 'guid'):
                    fileInfo.user_dn = getattr(self.step, "userDN", None)
                    fileInfo.async_dest = getattr(self.step, "asyncDest", None)
                    fileInfo.user_vogroup = getattr(self.step, "owner_vogroup", '')
                    fileInfo.user_vorole = getattr(self.step, "owner_vorole", '')

            stepReport.persist(reportLocation)

        return None

    # Accessory methods
    def handleLFNForMerge(self, mergefile, step):
        """
        _handleLFNForMerge_

        Digs up unmerged LFN out of WMStep outputModule and
        changes the current file to match.
        Requires a mergedLFNBase in the WMSpec output module
        """

        # First get the output module
        # Do this by finding the name in the step report
        # And then finding that module in the WMStep Helper

        outputName = getattr(mergefile, 'module_label', None)
        if not outputName:
            logging.error("Attempt to merge directly failed due to " \
                          + "No module_label in file.")
        if outputName.lower() == "merged":
            # Don't skip merge for merged files!
            return mergefile
        stepHelper = self.task.getStep(stepName=step)
        outputMod = stepHelper.getOutputModule(moduleName=outputName)

        if not outputMod:
            # Then we couldn't get the output module
            logging.error("Attempt to directly merge failed " \
                          + "due to no output module %s in WMStep" \
                          % (outputName))
            return mergefile

        # Okay, now we should have the output Module
        # Now we just need a second LFN

        newBase = getattr(outputMod, 'mergedLFNBase', None)
        oldBase = getattr(outputMod, 'lfnBase', None)

        if not newBase:
            # Then we don't have a base to change it to
            logging.error("Direct to Merge failed due to no mergedLFNBase in %s", outputName)
            return mergefile

        # Replace the actual LFN base
        oldLFN = getattr(mergefile, 'lfn')
        newLFN = oldLFN.replace(oldBase, newBase)

        # Set the file attributes
        setattr(mergefile, 'lfn', newLFN)
        setattr(mergefile, 'merged', True)

        # Return the file
        return mergefile



#!/usr/bin/env python
"""
_ParentlessMergeBySize_

WMBS merging that ignores file parents.
"""
from __future__ import division

import time
import threading
from functools import cmp_to_key

from WMCore.WMBS.File import File
from WMCore.DataStructs.Run import Run

from WMCore.DAOFactory import DAOFactory
from WMCore.JobSplitting.JobFactory import JobFactory

def fileCompare(a, b):
    """
    _fileCompare_

    Compare two files based on their "file_first_event" attribute.
    """
    if a["file_run"] > b["file_run"]:
        return 1
    elif a["file_run"] == b["file_run"]:
        if a["file_lumi"] > b["file_lumi"]:
            return 1
        elif a["file_lumi"] == b["file_lumi"]:
            if a["file_first_event"] > b["file_first_event"]:
                return 1
            if a["file_first_event"] == b["file_first_event"]:
                return 0

    return -1

class ParentlessMergeBySize(JobFactory):
    def defineFileGroups(self, mergeableFiles):
        """
        _defineFileGroups_

        Group mergeable files by their SE name and run number so that we don't
        try to merge together files on different SEs.  Merging against across
        run boundaries is configurable.
        """
        fileGroups = {}
        foundFiles = []

        for mergeableFile in mergeableFiles:
            if mergeableFile["file_lfn"] not in foundFiles:
                foundFiles.append(mergeableFile["file_lfn"])
            else:
                continue

            if mergeableFile["pnn"] not in fileGroups:
                if self.mergeAcrossRuns:
                    fileGroups[mergeableFile["pnn"]] = []
                else:
                    fileGroups[mergeableFile["pnn"]] = {}

            if self.mergeAcrossRuns:
                fileGroups[mergeableFile["pnn"]].append(mergeableFile)
            else:
                if mergeableFile["file_run"] not in fileGroups[mergeableFile["pnn"]]:
                    fileGroups[mergeableFile["pnn"]][mergeableFile["file_run"]] = []
                fileGroups[mergeableFile["pnn"]][mergeableFile["file_run"]].append(mergeableFile)

        return fileGroups

    def createMergeJob(self, mergeFiles):
        """
        _createMergeJob_

        Create a merge job for the given merge units.  All the files contained
        in the merge units will be associated to the job.
        """
        if self.currentGroup == None:
            self.newGroup()

        self.newJob(name = self.getJobName())
        mergeFiles.sort(key=cmp_to_key(fileCompare))

        jobSize = 0
        largestFile = 0
        for mergeFile in mergeFiles:

            jobSize += mergeFile["file_size"]
            largestFile = max(largestFile, mergeFile["file_size"])

            newFile = File(id = mergeFile["file_id"],
                           lfn = mergeFile["file_lfn"],
                           events = mergeFile["file_events"])

            # The WMBS data structure puts locations that are passed in through
            # the constructor in the "newlocations" attribute.  We want these to
            # be in the "locations" attribute so that they get picked up by the
            # job submitter.
            newFile["locations"] = set([mergeFile["pnn"]])
            newFile.addRun(Run(mergeFile["file_run"], mergeFile["file_lumi"]))
            self.currentJob.addFile(newFile)

        # job time based on
        #   - 5 min initialization
        #   - 5MB/s merge speed
        #   - checksum calculation at 5MB/s (twice)
        #   - stageout at 5MB/s
        # job disk based on
        #  - input for largest file on local disk
        #  - output on local disk (factor 1)
        jobTime = 300 + (jobSize*4)/5000000
        self.currentJob.addResourceEstimates(jobTime = jobTime, disk = (jobSize+largestFile)/1024)

        return

    def defineMergeJobs(self, mergeableFiles):
        """
        _defineMergeJobs_

        Try to define merge jobs that meet the size requirement.
        """
        mergeJobFileSize = 0
        mergeJobEvents   = 0
        mergeJobFiles    = []
        earliestInsert   = 999999999999999

        mergeableFiles.sort(key=cmp_to_key(fileCompare))

        for mergeableFile in mergeableFiles:
            if mergeableFile["file_size"] > self.maxMergeSize or \
                   mergeableFile["file_events"] > self.maxMergeEvents:
                self.createMergeJob([mergeableFile])
                continue
            elif mergeableFile["file_size"] + mergeJobFileSize > self.maxMergeSize or \
                     mergeableFile["file_events"] + mergeJobEvents > self.maxMergeEvents:
                if mergeJobFileSize > self.minMergeSize or \
                       self.forceMerge == True or \
                       time.time() - mergeableFile['insert_time'] > self.maxWaitTime:
                    self.createMergeJob(mergeJobFiles)
                    mergeJobFileSize = 0
                    mergeJobEvents = 0
                    mergeJobFiles = []
                else:
                    continue

            mergeJobFiles.append(mergeableFile)
            mergeJobFileSize += mergeableFile["file_size"]
            mergeJobEvents += mergeableFile["file_events"]
            if mergeableFile['insert_time'] < earliestInsert:
                earliestInsert = mergeableFile['insert_time']

        if mergeJobFileSize > self.minMergeSize or self.forceMerge == True or \
               time.time() - earliestInsert > self.maxWaitTime:
            if len(mergeJobFiles) > 0:
                self.createMergeJob(mergeJobFiles)

        return

    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Try to merge any available files for the subscription provided.  This
        accepts the following keyword arguments:
          max_merge_size - The maximum size of merged files
          min_merge_size - The minimum size of merged files
          max_merge_events - The maximum number of events in a merge file
          merge_across_runs - Whether or not we merge across runs
        """
        # This doesn't use a proxy
        self.grabByProxy = False

        self.maxMergeSize    = int(kwargs.get("max_merge_size", 1000000000))
        self.minMergeSize    = int(kwargs.get("min_merge_size", 1048576))
        self.maxMergeEvents  = int(kwargs.get("max_merge_events", 100000000))
        self.mergeAcrossRuns = kwargs.get("merge_across_runs", True)
        self.maxWaitTime     = kwargs.get("max_wait_time", 12 * 3600)

        self.subscription["fileset"].load()

        if self.subscription["fileset"].open == True:
            self.forceMerge = False
        else:
            self.forceMerge = True

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package = "WMCore.WMBS",
                                logger = myThread.logger,
                                dbinterface = myThread.dbi)

        if self.forceMerge:
            # Check and see if we have an injected workflow
            # If the workflow is not injected, then we can't forceMerge
            getAction = daoFactory(classname = "Workflow.CheckInjectedWorkflow")
            injected  = getAction.execute(name = self.subscription["workflow"].name,
                                          conn = myThread.transaction.conn,
                                          transaction = True)

            if not injected:
                self.forceMerge = False

        mergeDAO       = daoFactory(classname = "Subscriptions.GetFilesForParentlessMerge")
        mergeableFiles = mergeDAO.execute(self.subscription["id"],
                                          conn = myThread.transaction.conn,
                                          transaction = True)

        groupedFiles = self.defineFileGroups(mergeableFiles)

        for pnn in groupedFiles:
            if self.mergeAcrossRuns:
                self.defineMergeJobs(groupedFiles[pnn])
            else:
                for runNumber in groupedFiles[pnn]:
                    self.defineMergeJobs(groupedFiles[pnn][runNumber])

        return



#!/usr/bin/env python
"""
Component responsible for locally updating workflow specs and sandboxes.
"""

import logging
import threading
from pprint import pformat

from WMComponent.WorkflowUpdater.WorkflowUpdaterPoller import WorkflowUpdaterPoller
from WMCore.Agent.Harness import Harness


class WorkflowUpdater(Harness):
    """
    Create a WorkflowUpdaterPoller component as a daemon
    """

    def __init__(self, config):
        """
        Initialize it with the agent configuration parameters.
        :param config: a Configuration object with the component configuration
        """
        # call the base class
        Harness.__init__(self, config)

    def preInitialization(self):
        """
        Step that actually adds the worker thread properly
        """
        pollInterval = self.config.WorkflowUpdater.pollInterval
        logging.info("Starting %s with configuration:\n%s", self.__class__.__name__,
                     pformat(self.config.WorkflowUpdaterPoller.dictionary_()))

        myThread = threading.currentThread()
        myThread.workerThreadManager.addWorker(WorkflowUpdaterPoller(self.config),
                                               pollInterval)



#!/usr/bin/env python
"""
_ListAlgoDatasetAssoc_

MySQL implementation of DBSBuffer.ListAlgoDatasetAssoc
"""




from WMCore.Database.DBFormatter import DBFormatter

class ListAlgoDatasetAssoc(DBFormatter):
    """
    _ListDataset_

    Retrieve information about a dataset/algorthim association in the DBSBuffer.
    This is mostly used by the unit tests.
    """
    sql = """SELECT id, algo_id, dataset_id, in_dbs FROM dbsbuffer_algo_dataset_assoc
               WHERE id = :id"""

    def execute(self, assocID = None, conn = None, transaction = False):
        """
        _execute_

        Retrieve information about a dataset/algorithm association in DBSBuffer.
        """
        binds = {"id": assocID}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                          transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
Oracle implementation of Jobs.LoadOutputID

Load the ID of the output fileset for a job
"""
__all__ = []



from WMCore.WMBS.MySQL.Jobs.LoadOutputID import LoadOutputID as MySQLLoadOutputID

class LoadOutputID(MySQLLoadOutputID):
    """

    Load the ID of the output fileset for a job

    """



#!/usr/bin/env python
"""
_GetPNNtoPSNMapping_

MySQL implementation of Locations.GetPNNtoPSNMapping
"""
from __future__ import division
from __future__ import print_function

from WMCore.Database.DBFormatter import DBFormatter

__all__ = []


class GetPNNtoPSNMapping(DBFormatter):
    """
    Return a python dict mapping PhEDEx node name to list of related
    processing site names.
    """

    sql = """SELECT wpnn.pnn AS pnn, wl.site_name AS psn FROM wmbs_location_pnns wlpnn
             INNER JOIN wmbs_location wl ON wlpnn.location = wl.id
             INNER JOIN wmbs_pnns wpnn ON wlpnn.pnn = wpnn.id"""

    def execute(self, conn=None, transaction=False):
        results = self.dbi.processData(self.sql, conn=conn, transaction=transaction)
        mapping = {}
        for result in self.formatDict(results):
            mapping.setdefault(result['pnn'], []).append(result['psn'])
        return mapping



"""
MySQL implementation of Files.InFileset
"""
from WMCore.Database.DBFormatter import DBFormatter

class InFileset(DBFormatter):
    sql = "SELECT DISTINCT fileid AS fileid FROM wmbs_fileset_files WHERE fileset = :fileset"

    def execute(self, fileset = None, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"fileset": fileset},
                         conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_FileInfo_

Library for finding the information of a file
Primarily intended for use with jobReport, adding info
Not in the Framework XML


"""
from __future__ import print_function


from builtins import str, object

import os
import os.path
import logging

from Utils.FileTools import calculateChecksums


class FileInfo(object):
    """

    Package to hold all info regarding file

    """


    def __init__(self):
        """
        Doesn't do nothing

        """

    def __call__(self, fileReport, step, outputModule):
        """
        This takes a file from the report as an argument
        And extracts everything from there

        """

        pfn = fileReport.pfn
        if pfn.startswith("file:"):
            pfn = pfn.replace("file:", "")

        # First, where the hell is the file
        if not os.path.isfile(pfn):
            msg = "Fed non-valid PFN: %s" %(pfn)
            # Then figure out what to do with it
            # For now, do nothing since we can't find it
            logging.error(msg)
            print(msg)
            return
        if not os.path.isabs(pfn):
            # Get an absolute path
            pfn = os.path.abspath(pfn)


        # Now we know it, we better set it
        setattr(fileReport, 'pfn', pfn)

        # Set default of merged as false
        setattr(fileReport, 'merged', False)

        return self.processFile(filename = pfn,
                                fileReport = fileReport,
                                step = step,
                                outputModule = outputModule)




    def processFile(self, filename, fileReport, step, outputModule):
        """
        This attaches the info to the fileReport instance


        """
        # Get checksum
        (adler32, cksum) = calculateChecksums(filename)

        # Get info from spec
        output = getattr(step.output.modules, outputModule)
        disableGUID      = getattr(output, 'disableGUID', False)
        fixedLFN         = getattr(output, 'fixedLFN', False)
        primaryDataset   = output.primaryDataset
        processedDataset = output.processedDataset
        dataTier         = output.dataTier

        # Get other file information
        size = os.stat(filename)[6]

        #Get info from file
        mergedLFNBase    = getattr(fileReport, 'MergedLFNBase', None)
        mergedBySize     = getattr(fileReport, 'MergedBySize', False)
        lfn              = getattr(fileReport, 'lfn')

        # Do LFN manipulation
        # First in the standard case
        if not fixedLFN and not disableGUID:
            guid = getattr(fileReport, 'guid', None)
            if not guid:
                msg = "No GUID for file %s" %(lfn)
                logging.error(msg)
                raise Exception(msg)
            # Then we have to change the LFN to match the GUID
            dirname = os.path.dirname(lfn)
            filelfn = '%s.root' %(str(guid))
            setattr(fileReport, 'lfn', os.path.join(dirname, filelfn))
        elif not fixedLFN and mergedBySize and mergedLFNBase:
            # Then we better do the merge stuff
            # Not tested for now
            mergedLFNBase.rstrip('/')
            newLFN = os.path.join(mergedLFNBase, os.path.basename(lfn))
            setattr(fileReport, 'lfn', newLFN)




        # Attach values
        setattr(fileReport, 'checksums', {'adler32': adler32, 'cksum': cksum})
        setattr(fileReport, 'size', size)
        setattr(fileReport, "dataset", {"applicationName": "cmsRun",
                                        "applicationVersion": step.application.setup.cmsswVersion,
                                        "primaryDataset": primaryDataset,
                                        "processedDataset": processedDataset,
                                        "dataTier": dataTier})

        return fileReport



#!/usr/bin/env python
"""
Oracle implementation of AddFile
"""





from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.Add import Add as MySQLAdd

class Add(MySQLAdd):
    """
    Oracle implementation of AddFile
    """
    sql = """INSERT /*+ IGNORE_ROW_ON_DUPKEY_INDEX (dbsbuffer_file (lfn)) */ 
                 INTO dbsbuffer_file (lfn, filesize, events, dataset_algo, status, workflow, in_phedex)
                      values (:lfn, :filesize, :events, :dataset_algo, :status, :workflow, :in_phedex)"""



#!/usr/bin/env python
"""
_InsertSpec_

MySQL implementation of DBSBuffer3.UpdateSpec

Created on Mar 11, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class UpdateSpec(DBFormatter):
    """
    _UpdateSpec_

    Insert a spec using the name and task to a pre-existing record in dbsbuffer_workflow
    """

    existsSQL = "SELECT id FROM dbsbuffer_workflow WHERE name = :name AND task = :task"

    updateSQL = "UPDATE dbsbuffer_workflow SET spec = :spec WHERE id = :id"

    def execute(self, requestName, taskPath,
                specPath, conn = None, transaction = False):
        """
        _execute_

        Update a workflow in the dbsbuffer_workflow table
        """
        binds = {'name': requestName, 'task': taskPath}

        result = self.dbi.processData(self.existsSQL, binds, conn = conn,
                                      transaction = transaction)
        formattedResult = self.formatDict(result)

        if not formattedResult:
            return None

        workflowId = formattedResult[0]['id']

        if specPath is not None:
            binds = {'spec' : specPath, 'id' : workflowId}
            self.dbi.processData(self.updateSQL, binds, conn = conn,
                                 transaction = transaction)

        return workflowId



#!/usr/bin/env python
"""
_LCGImpl_

Implementation of StageOutImplV2 interface for lcg-cp

"""
import os, os.path, re, logging, subprocess, tempfile
from subprocess import Popen

from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from WMCore.Storage.StageOutError import StageOutError, StageOutFailure
from WMCore.Storage.Execute import runCommandWithOutput as runCommand

_CheckExitCodeOption = True



class LCGImpl(StageOutImplV2):
    """
    _LCGImpl_

    Implement interface for srmcp v2 command with lcg-* commands

    """

    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum ):
        """
            performs a transfer. stageOut tells you which way to go. returns the new pfn or
            raises on failure. StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin
        """
        localFileName = fromPfn
        if stageOut:
            fromPfn2 = self.prependFileProtocol(fromPfn)
            toPfn2 = toPfn
            localFileName = fromPfn
            remoteFileName = toPfn
        else:
            fromPfn2 = fromPfn
            toPfn2 = self.prependFileProtocol(toPfn)
            localFileName = toPfn
            remoteFileName = fromPfn
            localDir = os.path.dirname( localFileName )
            if not os.path.exists( localDir ):
                logging.info("Making local directory %s" % localDir)
                os.makedirs( localDir )

        if not options:
            options = ""

        transferCommand = "lcg-cp -b -D srmv2 --vo cms --srm-timeout 2400 --sendreceive-timeout 2400 --connect-timeout 300 --verbose %s %s %s " %\
                            ( options, fromPfn2, toPfn2 )

        logging.info("Staging out with lcg-cp")
        logging.info("  commandline: %s" % transferCommand)
        self.runCommandFailOnNonZero( transferCommand )

        logging.info("Verifying file sizes")
        localSize  = os.path.getsize( localFileName )
        remoteSize = subprocess.Popen(['lcg-ls', '-l', '-b', '-D', 'srmv2', remoteFileName],
                                       stdout=subprocess.PIPE).communicate()[0]
        logging.info("got the following from lcg-ls %s" % remoteSize)
        remoteSize = remoteSize.split()[4]
        logging.info("Localsize: %s Remotesize: %s" % (localSize, remoteSize))
        if int(localSize) != int(remoteSize):
            try:
                logging.error("Transfer failed, deleting partial file")
                self.doDelete(toPfn,None,None,None,None)
            except:
                pass
            raise StageOutFailure("File sizes don't match")


        return toPfn


    def doDelete(self, pfn, pnn, command, options, protocol  ):
        """
            deletes a file, raises on error
            StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin
        """
        if pfn.startswith("srm://"):
            runCommand( "lcg-del -b -l -D srmv2 --vo cms %s" % pfn )
        elif pfn.startswith("file:"):
            runCommand( "/bin/rm -f %s" % pfn.replace("file:", "", 1) )
        else:
            runCommand( StageOutImpl.createRemoveFileCommand(self, pfn) )



#!/usr/bin/env python
"""
_FailInput_

MySQL implementation of Jobs.FailInput
"""




from WMCore.Database.DBFormatter import DBFormatter

class FailInput(DBFormatter):
    """
    _FailInput_

    Given a job ID mark all it's input as failed for the subscription that
    produced it.  This will clear out any reference to the input files in the
    wmbs_sub_files_acquired and wmbs_sub_files_complete tables.
    """
    fileSelect = """SELECT wmbs_jobgroup.subscription AS subid,
                           wmbs_job_assoc.fileid AS fileid FROM wmbs_job_assoc
                      INNER JOIN wmbs_job ON
                        wmbs_job_assoc.job = wmbs_job.id
                      INNER JOIN wmbs_jobgroup ON
                        wmbs_job.jobgroup = wmbs_jobgroup.id
                      LEFT OUTER JOIN wmbs_sub_files_failed ON
                        wmbs_jobgroup.subscription = wmbs_sub_files_failed.subscription AND
                        wmbs_job_assoc.fileid = wmbs_sub_files_failed.fileid
                    WHERE wmbs_job.id = :jobid AND wmbs_sub_files_failed.fileid IS Null"""

    acquiredDelete = """DELETE FROM wmbs_sub_files_acquired
                        WHERE subscription = :subid AND fileid = :fileid"""

    completeDelete = """DELETE FROM wmbs_sub_files_complete
                      WHERE subscription = :subid AND fileid = :fileid"""

    sql = """INSERT IGNORE INTO wmbs_sub_files_failed (fileid, subscription)
               VALUES (:fileid, :subid)"""

    def execute(self, id, conn = None, transaction = False):
        results = self.dbi.processData(self.fileSelect, {"jobid": id}, conn = conn,
                                       transaction = transaction)
        delBinds = self.formatDict(results)

        if len(delBinds) > 0:
            self.dbi.processData(self.acquiredDelete, delBinds, conn = conn,
                                 transaction = transaction)
            self.dbi.processData(self.completeDelete, delBinds, conn = conn,
                                 transaction = transaction)
            self.dbi.processData(self.sql, delBinds, conn = conn,
                                 transaction = transaction)
        return



"""
_GetStateID_

MySQL implementation of Jobs.GetStateID

Created on Sep 13, 2012

@author: dballest
"""
from WMCore.Database.DBFormatter import DBFormatter

class GetStateID(DBFormatter):
    """
    DAO to retrieve the id of a given job state name
    """

    sql = """SELECT id
             FROM wmbs_job_state
             WHERE name = :stateName
          """

    def execute(self, state = 'created', conn = None, transaction = False):

        result = self.dbi.processData(self.sql, {'stateName' : state},
                             conn = conn, transaction = transaction)

        stateId = self.formatOne(result)

        return stateId[0]



#!/usr/bin/env python
"""
_GetJobGroups_

Oracle implementation of Subscription.GetJobGroups
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter
from WMCore.WMBS.MySQL.Jobs.ChangeState import ChangeState as BaseDAO

class ChangeState(BaseDAO):
    pass



from __future__ import print_function

import gzip
from builtins import str, bytes, object

from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytes, encodeUnicodeToBytesConditional
from future.utils import viewitems

import hashlib
import json
import xml.sax.saxutils
import zlib
from traceback import format_exc

import cherrypy

from WMCore.REST.Error import RESTError, ExecutionError, report_rest_error

try:
    from cherrypy.lib import httputil
except ImportError:
    from cherrypy.lib import http as httputil

def vary_by(header):
    """Add 'Vary' header for `header`."""
    varies = cherrypy.response.headers.get('Vary', '')
    varies = [x.strip() for x in varies.split(",") if x.strip()]
    if header not in varies:
        varies.append(header)
    cherrypy.response.headers['Vary'] = ", ".join(varies)

def is_iterable(obj):
    """Check if `obj` is iterable."""
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True

class RESTFormat(object):
    def __call__(self, stream, etag):
        """Main entry point for generating output for `stream` using `etag`
        object to generate ETag header. Returns a generator function for
        producing a verbatim copy of `stream` item, including any premables
        and trailers needed for the selected format. The intention is that
        the caller will use the iterable to generate chunked HTTP transfer
        encoding, or a simple result such as an image."""
        # Make 'stream' iterable. We convert everything to chunks here.
        # The final stream consumer will collapse small responses back
        # to a single string. Convert files to 1MB chunks.
        if stream is None:
            stream = ['']
        elif isinstance(stream, (str, bytes)):
            stream = [stream]
        elif hasattr(stream, "read"):
            # types.FileType is not available anymore in python3,
            # using it raises pylint W1624.
            # Since cherrypy.lib.file_generator only uses the .read() attribute
            # of a file, we simply check if stream.read() is present instead.
            # https://github.com/cherrypy/cherrypy/blob/2a8aaccd649eb1011382c39f5cd93f76f980c0b1/cherrypy/lib/__init__.py#L64
            stream = cherrypy.lib.file_generator(stream, 512 * 1024)

        return self.stream_chunked(stream, etag, *self.chunk_args(stream))

    def chunk_args(self, stream):
        """Return extra arguments needed for `stream_chunked()`. The default
        return an empty tuple, so no extra arguments. Override in the derived
        class if `stream_chunked()` needs preamble or trailer arguments."""
        return tuple()

class XMLFormat(RESTFormat):
    """Format an iterable of objects into XML encoded in UTF-8.

    Generates normally first a preamble, a stream of XML-rendered objects,
    then the trailer, computing an ETag on the output string in the process.
    This is designed exclusively for use with iterables for chunked transfer
    encoding HTTP responses; it's not a general purpose formatting utility.

    Outputs first a preamble, then XML encoded output of input stream, and
    finally a trailer. Any exceptions raised by input stream are reported to
    `report_rest_error` and swallowed, as this is normally used to generate
    output for CherryPy responses, which cannot handle exceptions reasonably
    after the output generation begins; later processing may reconvert those
    back to exceptions however (cf. stream_maybe_etag()). Once the preamble
    has been emitted, the trailer is also emitted even if the input stream
    raises an exception, in order to make the output well-formed; the client
    must inspect the X-REST-Status trailer header to find out if it got the
    complete output. No ETag header is generated in case of an exception.

    The ETag generation is deterministic only if iterating over input is
    deterministic. Beware in particular the key order for a dict is
    arbitrary and may differ for two semantically identical dicts.

    A X-REST-Status trailer header is added only in case of error. There is
    normally 'X-REST-Status: 100' in normal response headers, and it remains
    valid in case of success.

    The output is generated as an XML document whose top-level entity name
    is defined by the label given at the formatter construction time. The
    caller must define ``cherrypy.request.rest_generate_data`` to element
    name for wrapping stream contents. Usually the top-level entity is the
    application name and the ``cherrypy.request.rest_generate_data`` is
    ``result``.

    Iterables are output as ``<array><i>ITEM</i><i>ITEM</i></array>``,
    dictionaries as ``<dict><key>KEY</key><value>VALUE</value></dict>``.
    `None` is output as empty contents, and hence there is no way to
    distinguish `None` and an empty string from each other. Scalar types
    are output as rendered by `str()`, but obviously XML encoding unsafe
    characters. This class does not support formatting arbitrary types.

    The formatter does not insert any spaces into the output. Although the
    output is generated as a preamble, stream of objects, and trailer just
    like by the `JSONFormatter`, each of which is a separate HTTP transfer
    chunk, the output does *not* have guaranteed line-oriented structure
    like the `JSONFormatter` produces. Note in particular that if the data
    stream contains strings with newlines, the output will have arbitrary
    line structure. On the other hand, as the output is well-formed XML,
    virtually all SAX processors can read the stream incrementally even if
    the client isn't able to fully preserve chunked HTTP transfer encoding."""

    def __init__(self, label):
        self.label = label

    @staticmethod
    def format_obj(obj):
        """Render an object `obj` into XML."""
        if isinstance(obj, type(None)):
            result = ""
        elif isinstance(obj, str):
            result = xml.sax.saxutils.escape(obj).encode("utf-8")
        elif isinstance(obj, bytes):
            result = xml.sax.saxutils.escape(obj)
        elif isinstance(obj, (int, float, bool)):
            result = xml.sax.saxutils.escape(str(obj)).encode("utf-8")
        elif isinstance(obj, dict):
            result = "<dict>"
            for k, v in viewitems(obj):
                result += "<key>%s</key><value>%s</value>" % \
                  (xml.sax.saxutils.escape(k).encode("utf-8"),
                   XMLFormat.format_obj(v))
            result += "</dict>"
        elif is_iterable(obj):
            result = "<array>"
            for v in obj:
                result += "<i>%s</i>" % XMLFormat.format_obj(v)
            result += "</array>"
        else:
            cherrypy.log("cannot represent object of type %s in xml (%s)"
                         % (type(obj).__class__.__name__, repr(obj)))
            raise ExecutionError("cannot represent object in xml")
        return result

    def stream_chunked(self, stream, etag, preamble, trailer):
        """Generator for actually producing the output."""
        try:
            etag.update(preamble)
            yield preamble

            try:
                for obj in stream:
                    chunk = XMLFormat.format_obj(obj)
                    etag.update(chunk)
                    yield chunk
            except GeneratorExit:
                etag.invalidate()
                trailer = None
                raise
            finally:
                if trailer:
                    etag.update(trailer)
                    yield trailer

        except RESTError as e:
            etag.invalidate()
            report_rest_error(e, format_exc(), False)
        except Exception as e:
            etag.invalidate()
            report_rest_error(ExecutionError(), format_exc(), False)

    def chunk_args(self, stream):
        """Return header and trailer needed to wrap `stream` as XML reply."""
        preamble = "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
        preamble += "<%s>" % self.label
        if cherrypy.request.rest_generate_preamble:
            desc = self.format_obj(cherrypy.request.rest_generate_preamble)
            preamble += "<desc>%s</desc>" % desc
        preamble += "<%s>" % cherrypy.request.rest_generate_data
        trailer = "</%s></%s>" % (cherrypy.request.rest_generate_data, self.label)
        return preamble, trailer

class JSONFormat(RESTFormat):
    """Format an iterable of objects into JSON.

    Generates normally first a preamble, a stream of JSON-rendered objects,
    then the trailer, computing an ETag on the output string in the process.
    This is designed exclusively for use with iterables for chunked transfer
    encoding HTTP responses; it's not a general purpose formatting utility.

    Outputs first a preamble, then JSON encoded output of input stream, and
    finally a trailer. Any exceptions raised by input stream are reported to
    `report_rest_error` and swallowed, as this is normally used to generate
    output for CherryPy responses, which cannot handle exceptions reasonably
    after the output generation begins; later processing may reconvert those
    back to exceptions however (cf. stream_maybe_etag()). Once the preamble
    has been emitted, the trailer is also emitted even if the input stream
    raises an exception, in order to make the output well-formed; the client
    must inspect the X-REST-Status trailer header to find out if it got the
    complete output. No ETag header is generated in case of an exception.

    The ETag generation is deterministic only if `cjson.encode()` output is
    deterministic for the input. Beware in particular the key order for a
    dict is arbitrary and may differ for two semantically identical dicts.

    A X-REST-Status trailer header is added only in case of error. There is
    normally 'X-REST-Status: 100' in normal response headers, and it remains
    valid in case of success.

    The output is always generated as a JSON dictionary. The caller must
    define ``cherrypy.request.rest_generate_data`` as the key for actual
    contents, usually something like "result". The `stream` value will be
    generated as an array value for that key.

    If ``cherrypy.request.rest_generate_preamble`` is a non-empty list, it
    is output as the ``desc`` key value in the preamble before outputting
    the `stream` contents. Otherwise the output consists solely of `stream`.
    A common use of ``rest_generate_preamble`` is list of column labels
    with `stream` an iterable of lists of column values.

    The output is guaranteed to contain one line of preamble which starts a
    dictionary and an array ("``{key: [``"), one line of JSON rendering of
    each object in `stream`, with the first line starting with exactly one
    space and second and subsequent lines starting with a comma, and one
    final trailer line consisting of "``]}``". Each line is generated as a
    HTTP transfer chunk. This format is fixed so readers can be constructed
    to read and parse the stream incrementally one line at a time,
    facilitating maximum throughput processing of the response."""

    def stream_chunked(self, stream, etag, preamble, trailer):
        """Generator for actually producing the output."""
        comma = " "

        try:
            if preamble:
                etag.update(preamble)
                yield preamble

            obj = None
            try:
                for obj in stream:
                    chunk = comma + json.dumps(obj) + "\n"
                    etag.update(chunk)
                    yield chunk
                    comma = ","
            except cherrypy.HTTPError:
                raise
            except GeneratorExit:
                etag.invalidate()
                trailer = None
                raise
            except Exception as exp:
                print("ERROR, json.dumps failed to serialize %s, type %s\nException: %s" \
                        % (obj, type(obj), str(exp)))
                raise
            finally:
                if trailer:
                    etag.update(trailer)
                    yield trailer

            cherrypy.response.headers["X-REST-Status"] = 100
        except cherrypy.HTTPError:
            raise
        except RESTError as e:
            etag.invalidate()
            report_rest_error(e, format_exc(), False)
        except Exception as e:
            etag.invalidate()
            report_rest_error(ExecutionError(), format_exc(), False)

    def chunk_args(self, stream):
        """Return header and trailer needed to wrap `stream` as JSON reply."""
        comma = ""
        preamble = "{"
        trailer = "]}\n"
        if cherrypy.request.rest_generate_preamble:
            desc = json.dumps(cherrypy.request.rest_generate_preamble)
            preamble += '"desc": %s' % desc
            comma = ", "
        preamble += '%s"%s": [\n' % (comma, cherrypy.request.rest_generate_data)
        return preamble, trailer

class PrettyJSONFormat(JSONFormat):
    """ Format used for human, (web browser)"""

    def stream_chunked(self, stream, etag, preamble, trailer):
        """Generator for actually producing the output."""
        comma = " "

        try:
            if preamble:
                etag.update(preamble)
                yield preamble

            try:
                for obj in stream:
                    chunk = comma + json.dumps(obj, indent=2)
                    etag.update(chunk)
                    yield chunk
                    comma = ","
            except GeneratorExit:
                etag.invalidate()
                trailer = None
                raise
            finally:
                if trailer:
                    etag.update(trailer)
                    yield trailer

            cherrypy.response.headers["X-REST-Status"] = 100
        except RESTError as e:
            etag.invalidate()
            report_rest_error(e, format_exc(), False)
        except Exception as e:
            etag.invalidate()
            report_rest_error(ExecutionError(), format_exc(), False)

class PrettyJSONHTMLFormat(PrettyJSONFormat):
    """ Format used for human, (web browser) wrap around html tag on json"""

    @staticmethod
    def format_obj(obj):
        """Render an object `obj` into HTML."""
        if isinstance(obj, type(None)):
            result = ""
        elif isinstance(obj, str):
            obj = xml.sax.saxutils.quoteattr(obj)
            result = "<pre>%s</pre>" % obj if '\n' in obj else obj
        elif isinstance(obj, bytes):
            obj = xml.sax.saxutils.quoteattr(str(obj, "utf-8"))
            result = "<pre>%s</pre>" % obj if '\n' in obj else obj
        elif isinstance(obj, (int, float, bool)):
            result = "%s" % obj
        elif isinstance(obj, dict):
            result = "<ul>"
            for k, v in viewitems(obj):
                result += "<li><b>%s</b>: %s</li>" % (k, PrettyJSONHTMLFormat.format_obj(v))
            result += "</ul>"
        elif is_iterable(obj):
            empty = True
            result = "<details open><ul>"
            for v in obj:
                empty = False
                result += "<li>%s</li>" % PrettyJSONHTMLFormat.format_obj(v)
            result += "</ul></details>"
            if empty:
                result = ""
        else:
            cherrypy.log("cannot represent object of type %s in xml (%s)"
                         % (type(obj).__class__.__name__, repr(obj)))
            raise ExecutionError("cannot represent object in xml")
        return result

    def stream_chunked(self, stream, etag, preamble, trailer):
        """Generator for actually producing the output."""
        try:
            etag.update(preamble)
            yield preamble

            try:
                for obj in stream:
                    chunk = PrettyJSONHTMLFormat.format_obj(obj)
                    etag.update(chunk)
                    yield chunk
            except GeneratorExit:
                etag.invalidate()
                trailer = None
                raise
            finally:
                if trailer:
                    etag.update(trailer)
                    yield trailer

        except RESTError as e:
            etag.invalidate()
            report_rest_error(e, format_exc(), False)
        except Exception as e:
            etag.invalidate()
            report_rest_error(ExecutionError(), format_exc(), False)

    def chunk_args(self, stream):
        """Return header and trailer needed to wrap `stream` as XML reply."""
        preamble = "<html><body>"
        trailer = "</body></html>"
        return preamble, trailer

class RawFormat(RESTFormat):
    """Format an iterable of objects as raw data.

    Generates raw data completely unmodified, for example image data or
    streaming arbitrary external data files including even plain text.
    Computes an ETag on the output in the process. The result is always
    chunked, even simple strings on input. Usually small enough responses
    will automatically be converted back to a single string response post
    compression and ETag processing.

    Any exceptions raised by input stream are reported to `report_rest_error`
    and swallowed, as this is normally used to generate output for CherryPy
    responses, which cannot handle exceptions reasonably after the output
    generation begins; later processing may reconvert those back to exceptions
    however (cf. stream_maybe_etag()). A X-REST-Status trailer header is added
    if (and only if) an exception occurs; the client must inspect that to find
    out if it got the complete output. There is normally 'X-REST-Status: 100'
    in normal response headers, and it remains valid in case of success.
    No ETag header is generated in case of an exception."""

    def stream_chunked(self, stream, etag):
        """Generator for actually producing the output."""
        try:
            for chunk in stream:
                etag.update(chunk)
                yield chunk

        except RESTError as e:
            etag.invalidate()
            report_rest_error(e, format_exc(), False)
        except Exception as e:
            etag.invalidate()
            report_rest_error(ExecutionError(), format_exc(), False)
        except BaseException:
            etag.invalidate()
            raise

class DigestETag(object):
    """Compute hash digest over contents for ETag header."""
    algorithm = None

    def __init__(self, algorithm=None):
        """Prepare ETag computer."""
        self.digest = hashlib.new(algorithm or self.algorithm)

    def update(self, val):
        """Process response data `val`."""
        if self.digest:
            self.digest.update(encodeUnicodeToBytes(val))

    def value(self):
        """Return ETag header value for current input."""
        return self.digest and '"%s"' % self.digest.hexdigest()

    def invalidate(self):
        """Invalidate the ETag calculator so value() will return None."""
        self.digest = None

class MD5ETag(DigestETag):
    """Compute MD5 hash over contents for ETag header."""
    algorithm = 'md5'

class SHA1ETag(DigestETag):
    """Compute SHA1 hash over contents for ETag header."""
    algorithm = 'sha1'

def _stream_compress_identity(reply, *args):
    """Streaming compressor which returns original data unchanged."""
    return reply

def _stream_compress_deflate(reply, compress_level, max_chunk):
    """Streaming compressor for the 'deflate' method. Generates output that
    is guaranteed to expand at the exact same chunk boundaries as original
    reply stream."""

    # Create zlib compression object, with raw data stream (negative window size)
    z = zlib.compressobj(compress_level, zlib.DEFLATED, -zlib.MAX_WBITS,
                         zlib.DEF_MEM_LEVEL, 0)

    # Data pending compression. We only take entire chunks from original
    # reply. Then process reply one chunk at a time. Whenever we have enough
    # data to compress, spit it out flushing the zlib engine entirely, so we
    # respect original chunk boundaries.
    npending = 0
    pending = []
    for chunk in reply:
        pending.append(chunk)
        npending += len(chunk)
        if npending >= max_chunk:
            part = z.compress(encodeUnicodeToBytes("".join(pending))) + z.flush(zlib.Z_FULL_FLUSH)
            pending = []
            npending = 0
            yield part

    # Crank the compressor one more time for remaining output.
    if npending:
        yield z.compress(encodeUnicodeToBytes("".join(pending))) + z.flush(zlib.Z_FINISH)


def _stream_compress_gzip(reply, compress_level, *args):
    """Streaming compressor for the 'gzip' method. Generates output that
    is guaranteed to expand at the exact same chunk boundaries as original
    reply stream."""
    data = []
    for chunk in reply:
        data.append(chunk)
    if data:
        yield gzip.compress(encodeUnicodeToBytes("".join(data)), compress_level)


# : Stream compression methods.
_stream_compressor = {
  'identity': _stream_compress_identity,
  'deflate': _stream_compress_deflate,
  'gzip': _stream_compress_gzip
}

def stream_compress(reply, available, compress_level, max_chunk):
    """If compression has been requested via Accept-Encoding request header,
    and is granted for this response via `available` compression methods,
    convert the streaming `reply` into another streaming response which is
    compressed at the exact chunk boundaries of the original response,
    except that individual chunks may be coalesced up to `max_chunk` size.
    The `compression_level` tells how hard to compress, zero disables the
    compression entirely."""

    global _stream_compressor
    for enc in cherrypy.request.headers.elements('Accept-Encoding'):
        if enc.value not in available:
            continue

        elif enc.value in _stream_compressor and compress_level > 0:
            # Add 'Vary' header for 'Accept-Encoding'.
            vary_by('Accept-Encoding')

            # Compress contents at original chunk boundaries.
            if 'Content-Length' in cherrypy.response.headers:
                del cherrypy.response.headers['Content-Length']
            cherrypy.response.headers['Content-Encoding'] = enc.value
            return _stream_compressor[enc.value](reply, compress_level, max_chunk)

    return reply

def _etag_match(status, etagval, match, nomatch):
    """Match ETag value against any If-Match / If-None-Match headers."""
    # Execute conditions only for status 2xx. We only handle GET/HEAD
    # requests here, it makes no sense to try to do this for PUT etc.
    # as they need to be handled as request pre-condition, not in the
    # streaming out part here.
    if cherrypy.request.method in ('GET', 'HEAD'):
        status, dummyReason, dummyMsg = httputil.valid_status(status)
        if status >= 200 and status <= 299:
            if match and ("*" in match or etagval in match):
                raise cherrypy.HTTPError(412, "Precondition on ETag %s failed" % etagval)
            if nomatch and ("*" in nomatch or etagval in nomatch):
                raise cherrypy.HTTPRedirect([], 304)

def _etag_tail(head, tail, etag):
    """Generator which first returns anything in `head`, then `tail`.
    Sets ETag header at the end to value of `etag` if it's defined and
    yields a value."""
    for chunk in head:
        yield encodeUnicodeToBytes(chunk)

    for chunk in tail:
        yield encodeUnicodeToBytes(chunk)

    etagval = (etag and etag.value())
    if etagval:
        cherrypy.response.headers["ETag"] = etagval

def stream_maybe_etag(size_limit, etag, reply):
    """Maybe generate ETag header for the response, and handle If-Match
    and If-None-Match request headers. Consumes the reply until at most
    `size_limit` bytes. If the response fits into that size, adds the
    ETag header and matches it against any If-Match / If-None-Match
    request headers and replies appropriately.

    If the response is fully buffered, and the `reply` generator actually
    results in an error and sets X-Error-HTTP / X-Error-Detail headers,
    converts that error back into a real HTTP error response. Otherwise
    responds with the fully buffered body directly, without generator
    and chunking. In other words, responses smaller than `size_limit`
    are always fully buffered and replied immediately without chunking.
    If the response is not fully buffered, it's guaranteed to be output
    at original chunk boundaries.

    Note that if this function is fed the output from `stream_compress()`
    as it normally would be, the `size_limit` constrains the compressed
    size, and chunk boundaries correspond to compressed chunks."""

    req = cherrypy.request
    res = cherrypy.response
    match = [str(x) for x in (req.headers.elements('If-Match') or [])]
    nomatch = [str(x) for x in (req.headers.elements('If-None-Match') or [])]

    # If ETag is already set, match conditions and output without buffering.
    etagval = res.headers.get('ETag', None)
    if etagval:
        _etag_match(res.status or 200, etagval, match, nomatch)
        res.headers['Trailer'] = 'X-REST-Status'
        return _etag_tail([], reply, None)

    # Buffer up to size_limit bytes internally. This interally builds up the
    # ETag value inside 'etag'. In case of exceptions the ETag invalidates.
    # If we exceed the limit, fall back to streaming without checking ETag
    # against If-Match/If-None-Match. We'll still set the ETag in the trailer
    # headers, so clients which understand trailers will get the value; most
    # clients including browsers will ignore them.
    size = 0
    result = []
    for chunk in reply:
        result.append(chunk)
        size += len(chunk)
        if size > size_limit:
            res.headers['Trailer'] = 'X-REST-Status'
            return _etag_tail(result, reply, etag)

    # We've buffered the entire response, but it may be an error reply. The
    # generator code does not know if it's allowed to raise exceptions, so
    # it swallows all errors and converts them into X-* headers. We recover
    # the original HTTP response code and message from X-Error-{HTTP,Detail}
    # headers, if any are present.
    err = res.headers.get('X-Error-HTTP', None)
    if err:
        message = res.headers.get('X-Error-Detail', 'Original error lost')
        raise cherrypy.HTTPError(int(err), message)

    # OK, we buffered the entire reply and it's ok. Check ETag match criteria.
    # The original stream generator must guarantee that if it fails it resets
    # the 'etag' value, even if the error handlers above didn't run.
    etagval = etag.value()
    if etagval:
        res.headers['ETag'] = etagval
        _etag_match(res.status or 200, etagval, match, nomatch)

    # OK, respond with the buffered reply as a plain string.
    res.headers['Content-Length'] = size
    # TODO investigate why `result` is a list of bytes strings in py3
    # The current solution seems to work in both py2 and py3
    resp = b"" if PY3 else ""
    for item in result:
        resp += encodeUnicodeToBytesConditional(item, condition=PY3)
    assert len(resp) == size
    return resp



#!/usr/bin/env python
"""
_GetOutputMap_

MySQL implementation of Jobs.GetOutputMap
"""

from WMCore.Database.DBFormatter import DBFormatter

class GetOutputMap(DBFormatter):
    sql = """SELECT wmbs_workflow_output.output_identifier AS wf_output_id,
                    wmbs_workflow_output.output_fileset AS wf_output_fset,
                    wmbs_workflow_output.merged_output_fileset AS wf_output_mfset
                    FROM wmbs_workflow_output
               INNER JOIN wmbs_subscription ON
                 wmbs_workflow_output.workflow_id = wmbs_subscription.workflow
               INNER JOIN wmbs_jobgroup ON
                 wmbs_subscription.id = wmbs_jobgroup.subscription
               INNER JOIN wmbs_job ON
                 wmbs_jobgroup.id = wmbs_job.jobgroup
             WHERE wmbs_job.id = :jobid"""

    def execute(self, jobID, conn = None, transaction = False):
        results = self.dbi.processData(self.sql, {"jobid": jobID}, conn = conn,
                                       transaction = transaction)

        outputMap = {}
        for result in self.formatDict(results):
            if result["wf_output_id"] not in outputMap:
                outputMap[result["wf_output_id"]] = []

            outputMap[result["wf_output_id"]].append({"output_fileset": result["wf_output_fset"],
                                                      "merged_output_fileset": result["wf_output_mfset"]})
        return outputMap



#!/usr/bin/env python
"""
Perform cleanup actions
"""
import time
import random
import threading

from Utils.Timers import timeFunction
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.Services.ReqMgr.ReqMgr import ReqMgr
from WMCore.DAOFactory import DAOFactory


class WorkQueueManagerCleaner(BaseWorkerThread):
    """
    Cleans expired items, updates element status.
    """
    def __init__(self, queue, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)
        self.forbiddenStatus = ["aborted", "aborted-completed", "force-complete", "completed"]
        self.queue = queue
        self.config = config
        self.reqmgr2Svc = ReqMgr(self.config.General.ReqMgr2ServiceURL)
        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)
        self.finishedWorflowCheck = daoFactory(classname="Subscriptions.CountFinishedSubscriptionsByWorkflow")

    def setup(self, parameters):
        """
        Called at startup - introduce random delay
             to avoid workers all starting at once
        """
        t = random.randrange(self.idleTime)
        self.logger.info('Sleeping for %d seconds before 1st loop' % t)
        time.sleep(t)

    @timeFunction
    def algorithm(self, parameters):
        """
        Check & expire negotiation failures
        """
        self.queue.logger.info("Start updating & cleaning...")
        try:
            self.queue.performQueueCleanupActions()
            # this will clean up whatever left over from above clean up.
            # also if the wq replication has problem it won't delay the killing jobs in condor
            # and updating wmbs status
            # state lists which shouldn't be populated in wmbs. (To prevent creating work before WQE status updated)
            # added completed status in the list due to the race condition
            requests = self.reqmgr2Svc.getRequestByStatusFromMemoryCache(self.forbiddenStatus).getData()
            results = self.finishedWorflowCheck.execute(workflowNames=requests)

            requestsToKill = [reqInfo["workflow"] for reqInfo in results if reqInfo["open"] > 0]

            self.queue.logger.info("Killing %d requests in WMBS ...", len(requestsToKill))
            self.queue.killWMBSWorkflows(requestsToKill)
        except Exception as ex:
            self.queue.logger.exception("Error cleaning queue: %s", str(ex))

        self.queue.logger.info("Finished updating & cleaning.")



#!/bin/env python


"""
_LinearAlgo_

Straight lines to the end
"""

from WMComponent.RetryManager.PlugIns.RetryAlgoBase import RetryAlgoBase


class LinearAlgo(RetryAlgoBase):
    """
    _LinearAlgo_

    Make a linear backoff for the jobs
    """


    def isReady(self, job, cooloffType):
        """
        Actual function that does the work

        """

        # Get the cooloff time
        baseTimeoutDict = self.getAlgoParam(job['jobType'])
        baseTimeout = baseTimeoutDict.get(cooloffType.lower(), 10)
        cooloffTime = baseTimeout * job['retry_count']

        currentTime = self.timestamp()
        if currentTime - job['state_time'] > cooloffTime:
            return True
        else:
            return False



#!/usr/bin/env python
"""
UpdateWorkflowsToCompleted
"""

from WMCore.Database.DBFormatter import DBFormatter

class UpdateWorkflowsToCompleted(DBFormatter):
    """
    _UpdateWorkflosToCompleted_

    Update dbsbuffer workflow to complete
    """

    sql = """UPDATE dbsbuffer_workflow
             SET completed = 1
             WHERE name = :WORKFLOW
             """

    def execute(self, workflows, conn = None, transaction = False):
        """
        _execute_

        Run the query
        """
        binds = []
        for workflow in workflows:
            binds.append( { "WORKFLOW" : workflow } )

        self.dbi.processData(self.sql, binds,
                             conn = conn, transaction = transaction)

        return



#!/usr/bin/env python
"""
_GetBulkRunLumi_

MySQL implementation of GetBulkRunLumi
"""

from WMCore.Database.DBFormatter import DBFormatter


class GetBulkRunLumi(DBFormatter):
    """
    Note that this is ID based.  I may have to change it back
    to lfn based.

    """
    sql = """SELECT flr.run AS run, flr.lumi AS lumi, flr.fileid AS id
               FROM wmbs_file_runlumi_map flr
               WHERE flr.fileid = :id
    """

    def getBinds(self, files=None):
        binds = []
        files = self.dbi.makelist(files)
        for f in files:
            binds.append({'id': f['id']})
        return binds

    def format(self, result):
        "Return a list of Run/Lumi Set"

        finalResult = {}
        res = self.formatDict(result)

        for entry in res:
            fileid = entry['id']
            run = entry['run']
            finalResult.setdefault(fileid, {})
            finalResult[fileid].setdefault(run, [])
            finalResult[fileid][run].append(entry['lumi'])

        return finalResult

    def execute(self, files=None, conn=None, transaction=False):
        binds = self.getBinds(files)

        result = self.dbi.processData(self.sql, binds,
                                      conn=conn, transaction=transaction)
        return self.format(result)



#!/usr/bin/env python
"""
DeleteMgr

Util class to provide delete functionality as an interface object.

Based on StageOutMgr class

"""
from __future__ import print_function

from builtins import object
from future.utils import viewitems

import logging

from WMCore.Storage.Registry import retrieveStageOutImpl
# do we want seperate exceptions - for the moment no
from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutError import StageOutInitError
from WMCore.WMException import WMException


# from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig


class DeleteMgrError(WMException):
    """
    _DeleteMgrError_

    Specific exception class to work out file deletion exception details
    """

    def __init__(self, message, **data):
        WMException.__init__(self, message, **data)
        self.data.setdefault("ErrorCode", 60313)
        self.data.setdefault("ErrorType", self.__class__.__name__)


class DeleteMgr(object):
    """
    _DeleteMgr_

    Object that can be used to delete a set of files
    using TFC or an override.

    """

    def __init__(self, **overrideParams):
        self.override = False
        self.logger = overrideParams.pop("logger", logging.getLogger())
        self.overrideConf = overrideParams
        if overrideParams != {}:
            self.override = True

        #  //
        # // Try an get the TFC for the site
        # //
        self.tfc = None

        from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig

        self.numberOfRetries = 3
        self.retryPauseTime = 600
        self.pnn = None
        self.fallbacks = []

        #  //
        # // If override isnt None, we dont need SiteCfg, if it is
        # //  then we need siteCfg otherwise we are dead.

        if self.override == False:
            self.siteCfg = loadSiteLocalConfig()

        if self.override:
            self.initialiseOverride()
        else:
            self.initialiseSiteConf()

    def initialiseSiteConf(self):
        """
        _initialiseSiteConf_

        Extract required information from site conf and TFC

        """

        implName = self.siteCfg.localStageOut.get("command", None)
        if implName == None:
            msg = "Unable to retrieve local stage out command\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError(msg)
        msg = "Local Stage Out Implementation to be used is:"
        msg += "%s\n" % implName

        pnn = self.siteCfg.localStageOut.get("phedex-node", None)
        if pnn == None:
            msg = "Unable to retrieve local stage out phedex-node\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError(msg)
        msg += "Local Stage Out PNN to be used is %s\n" % pnn
        catalog = self.siteCfg.localStageOut.get("catalog", None)
        if catalog == None:
            msg = "Unable to retrieve local stage out catalog\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError(msg)
        msg += "Local Stage Out Catalog to be used is %s\n" % catalog

        try:
            self.tfc = self.siteCfg.trivialFileCatalog()
            msg += "Trivial File Catalog has been loaded:\n"
            msg += str(self.tfc)
        except Exception as ex:
            msg = "Unable to load Trivial File Catalog:\n"
            msg += "Local stage out will not be attempted\n"
            msg += str(ex)
            raise StageOutInitError(msg)

        self.logger.info(msg)
        self.pnn = pnn
        return

    def initialiseOverride(self):
        """
        _initialiseOverride_

        Extract and verify that the Override parameters are all present

        """
        overrideConf = self.overrideConf
        overrideParams = {
            "command": None,
            "option": None,
            "phedex-node": None,
            "lfn-prefix": None,
        }

        try:
            overrideParams['command'] = overrideConf['command']
            overrideParams['phedex-node'] = overrideConf['phedex-node']
            overrideParams['lfn-prefix'] = overrideConf['lfn-prefix']
        except Exception as ex:
            msg = "Unable to extract Override parameters from config:\n"
            msg += str(overrideConf)
            raise StageOutInitError(msg)
        if 'option' in overrideConf:
            if len(overrideConf['option']) > 0:
                overrideParams['option'] = overrideConf['option']
            else:
                overrideParams['option'] = ""

        msg = "=======Delete Override Initialised:================\n"
        for key, val in viewitems(overrideParams):
            msg += " %s : %s\n" % (key, val)
        msg += "=====================================================\n"
        self.logger.info(msg)
        self.fallbacks.append(overrideParams)
        self.pnn = overrideParams['phedex-node']
        return

    def __call__(self, fileToDelete):
        """
        _operator()_

        Use call to delete a file

        """
        self.logger.info("==>Working on file: %s" % fileToDelete['LFN'])

        lfn = fileToDelete['LFN']
        fileToDelete['PNN'] = self.pnn

        deleteSuccess = False

        #  //
        # // No override => use local-stage-out from site conf
        # //  invoke for all files and check failures/successes
        if not self.override:
            self.logger.info("===> Attempting To Delete.")
            try:
                fileToDelete['PFN'] = self.deleteLFN(lfn)
                deleteSuccess = True
            except Exception as ex:
                self.logger.error("===> Local file deletion failure. Exception:\n%s", str(ex))

        if not deleteSuccess and len(self.fallbacks) > 0:
            #  //
            # // Still here => override start using the fallback stage outs
            # //  If override is set, then that will be the only fallback available
            self.logger.info("===> Attempting To Delete files with fallback.")
            for fallback in self.fallbacks:
                if not deleteSuccess:
                    try:
                        fileToDelete['PFN'] = self.deleteLFN(lfn, fallback)
                        deleteSuccess = True
                    except Exception as ex:
                        continue

        if deleteSuccess:
            msg = "===> Delete Successful:\n"
            msg += "====> LFN: %s\n" % fileToDelete['LFN']
            msg += "====> PFN: %s\n" % fileToDelete['PFN']
            msg += "====> PNN:  %s\n" % fileToDelete['PNN']
            self.logger.info(msg)
            return fileToDelete
        else:
            msg = "Unable to delete file:\n"
            msg += fileToDelete['LFN']
            raise StageOutFailure(msg, **fileToDelete)

    def deleteLFN(self, lfn, override=None):
        """
        deleteLFN

        Given the lfn and local stage out params, invoke the delete

        if override is used the follwoing params should be defined:
        command - the stage out impl plugin name to be used
        option - the option values to be passed to that command (None is allowed)
        lfn-prefix - the LFN prefix to generate the PFN
        phedex-node - the Name of the PNN to which the file is being xferred
        """

        if override:
            command = override['command']
            pfn = "%s%s" % (override['lfn-prefix'], lfn)
        else:
            command = self.siteCfg.localStageOut['command']
            pfn = self.searchTFC(lfn)

        if pfn == None:
            msg = "Unable to match lfn to pfn: \n  %s" % lfn
            raise StageOutFailure(msg, LFN=lfn, TFC=str(self.tfc))

        return self.deletePFN(pfn, lfn, command)

    def deletePFN(self, pfn, lfn, command):
        """
        Delete the given PFN
        """
        try:
            impl = retrieveStageOutImpl(command)
        except Exception as ex:
            msg = "Unable to retrieve impl for file deletion in:\n"
            msg += "Error retrieving StageOutImpl for command named: %s\n" % (
                command,)
            raise StageOutFailure(msg, Command=command,
                                  LFN=lfn, ExceptionDetail=str(ex))
        impl.numRetries = self.numberOfRetries
        impl.retryPause = self.retryPauseTime

        try:
            impl.removeFile(pfn)
        except Exception as ex:
            self.logger.error("Failed to delete file: %s", pfn)
            ex.addInfo(Protocol=command, LFN=lfn, TargetPFN=pfn)
            raise ex

        return pfn

    #    def reportStageOutFailure(self, stageOutExcep):
    #        """
    #        _reportStageOutFailure_
    #
    #        When a stage out failure occurs, report it to the input
    #        framework job report.
    #
    #        - *stageOutExcep* : Instance of on of the StageOutError derived classes
    #
    #        """
    #        errStatus = stageOutExcep.data["ErrorCode"]
    #        errType = stageOutExcep.data["ErrorType"]
    #        desc = stageOutExcep.message
    #
    #        errReport = self.inputReport.addError(errStatus, errType)
    #        errReport['Description'] = desc
    #        return

    def searchTFC(self, lfn):
        """
        _searchTFC_

        Search the Trivial File Catalog for the lfn provided,
        if a match is made, return the matched PFN

        """
        if self.tfc == None:
            msg = "Trivial File Catalog not available to match LFN:\n"
            msg += lfn
            self.logger.error(msg)
            return None
        if self.tfc.preferredProtocol == None:
            msg = "Trivial File Catalog does not have a preferred protocol\n"
            msg += "which prevents local stage out for:\n"
            msg += lfn
            self.logger.error(msg)
            return None

        pfn = self.tfc.matchLFN(self.tfc.preferredProtocol, lfn)
        if pfn == None:
            msg = "Unable to map LFN to PFN:\n"
            msg += "LFN: %s\n" % lfn
            return None

        msg = "LFN to PFN match made:\n"
        msg += "LFN: %s\nPFN: %s\n" % (lfn, pfn)
        self.logger.info(msg)
        return pfn

##if __name__ == '__main__':
##    import StageOut.Impl
##    mgr = StageOutMgr()
##    pfn = "/home/evansde/work/PRODAGENT/work/JobCreator/RelValMinBias-170pre12/Processing/RelValMinBias-170pre12-Processing.tar.gz"
##    lfn = "/store/unmerged/mc/2007/11/13/RelVal-RelValMinBias-1194987281/GEN-SIM-DIGI-RECO/0201/DCCP-FNAL-TEST.dat"

##    mgr.searchTFC(lfn)
##    mgr(LFN = lfn, PFN = pfn, GUID=None)



#!/usr/bin/env python
"""
_LoadFromName_

Oracle implementation of Workflow.LoadFromName
"""

from WMCore.WMBS.MySQL.Workflow.LoadFromName import LoadFromName \
    as LoadFromNameMySQL


class LoadFromName(LoadFromNameMySQL):
    pass



#!/usr/bin/env python
"""
_SetLocation_

Oracle implementation of DBSBufferFiles.SetLocation
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.SetLocation import SetLocation as \
     MySQLSetLocation

class SetLocation(MySQLSetLocation):
    """
    Insert unique lfn, location row
    """
    sql = """INSERT INTO dbsbuffer_file_location (filename, location)
               SELECT :fileid, :locationid FROM DUAL WHERE NOT EXISTS
                 (SELECT filename FROM dbsbuffer_file_location
                 WHERE filename= :fileid and location= :locationid)
               """



#!/usr/bin/env python
"""
_AddValidation_

MySQL implementation of Subscription.AddValidation
"""

from WMCore.Database.DBFormatter import DBFormatter

class AddValidation(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_subscription_validation
                               (subscription_id, location_id, valid)
               SELECT :sub, id, :valid FROM wmbs_location
               WHERE site_name = :site_name"""

    def execute(self, sites, conn = None, transaction = False):
        self.dbi.processData(self.sql, sites, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_Utilities_

Common/useful utils.
Note that WMWorkload, WMTask and WMStep all depend on this module, so
dont import them in it.

"""


def findObjectTypeAboveNode(node, objType):
    """
    _findObjectTypeAboveNode_

    Given a config section (tree or not) traverse up the parent
    structure until finding the first entry containing an objectType
    setting that is set to the required type

    """
    if getattr(node, "objectType", None) == objType:
        return node
    if node._internal_parent_ref == None:
        return None
    return findObjectTypeAboveNode(node._internal_parent_ref, objType)


def findTaskAboveNode(node):
    """
    _findTaskAboveNode_

    Given a config section (tree or not) traverse up the parent
    structure until finding the first entry containing an objectType
    setting that is set to WMTask

    """
    return findObjectTypeAboveNode(node, "WMTask")

def findWorkloadAboveNode(node):
    """
    _findWorkloadAboveNode_

    Given a config section (tree or not) traverse up the parent
    structure until finding the first entry containing an objectType
    setting that is set to WMWorkload

    """
    return findObjectTypeAboveNode(node, "WMWorkload")



def stepIdentifier(stepHelper):
    """
    _stepIdentifier_

    Generate the /Workload/Task/Step id from the step instance
    given.

    """
    task = findTaskAboveNode(stepHelper.data)
    identifier = "%s/%s" % (task.pathName,
                            stepHelper.name())
    return identifier



#!/usr/bin/env python
"""
_GetAcquiredFiles_

MySQL implementation of Subscription.GetAcquiredFiles
"""

__all__ = []

from WMCore.WMBS.MySQL.Subscriptions.GetAvailableFiles import GetAvailableFiles


class GetAcquiredFiles(GetAvailableFiles):
    sql = """SELECT wmsfa.fileid, wl.site_name FROM wmbs_sub_files_acquired wmsfa
             INNER JOIN wmbs_file_location wfl ON wfl.fileid = wmsfa.fileid
             INNER JOIN wmbs_location_pnns wlpnn ON wlpnn.pnn = wfl.pnn
             INNER JOIN wmbs_location wl ON wl.id = wlpnn.location
             WHERE wmsfa.subscription = :subscription
             """

    def execute(self, subscription=None, conn=None, transaction=False):
        results = self.dbi.processData(self.sql, {"subscription": subscription},
                                       conn=conn, transaction=transaction)
        return self.formatDict(results)



#!/bin/env python
"""
Utilities related to timing and performance testing
"""

from builtins import object
import logging
import time
import calendar
from datetime import tzinfo, timedelta


def gmtimeSeconds():
    """
    Return GMT time in seconds
    """
    return int(time.mktime(time.gmtime()))


def encodeTimestamp(secs):
    """
    Encode second since epoch to a string GMT timezone representation
    :param secs: input timestamp value (either int or float) in seconds since epoch

    :return: time string in GMT timezone representation
    """
    if not isinstance(secs, (int, float)):
        raise Exception("Wrong input, should be seconds since epoch either int or float value")
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(int(secs)))


def decodeTimestamp(timeString):
    """
    Decode timestamps in provided document
    :param timeString: timestamp string represention in GMT timezone, see encodeTimestamp

    :return: seconds since ecouch in GMT timezone
    """
    if not isinstance(timeString, str):
        raise Exception("Wrong input, should be time string in GMT timezone representation")
    return calendar.timegm(time.strptime(timeString, "%Y-%m-%dT%H:%M:%SZ"))


def timeFunction(func):
    """
    source: https://www.andreas-jung.com/contents/a-python-decorator-for-measuring-the-execution-time-of-methods

    Decorator function to measure how long a method/function takes to run
    It returns a tuple with:
      * wall clock time spent
      * returned result of the function
      * the function name
    """

    def wrapper(*arg, **kw):
        t1 = time.time()
        res = func(*arg, **kw)
        t2 = time.time()
        return round((t2 - t1), 4), res, func.__name__

    return wrapper


class CodeTimer(object):
    """
    A context manager for timing function calls.
    Adapted from https://www.blog.pythonlibrary.org/2016/05/24/python-101-an-intro-to-benchmarking-your-code/

    Use like

    with CodeTimer(label='Doing something'):
        do_something()
    """

    def __init__(self, label='The function', logger=None):
        self.start = time.time()
        self.label = label
        self.logger = logger or logging.getLogger()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        end = time.time()
        runtime = round((end - self.start), 3)
        self.logger.info(f"{self.label} took {runtime} seconds to complete")


class LocalTimezone(tzinfo):
    """
    A required python 2 class to determine current timezone for formatting rfc3339 timestamps
    Required for sending alerts to the MONIT AlertManager
    Can be removed once WMCore starts using python3

    Details of class can be found at: https://docs.python.org/2/library/datetime.html#tzinfo-objects
    """

    def __init__(self):
        super(LocalTimezone, self).__init__()
        self.ZERO = timedelta(0)
        self.STDOFFSET = timedelta(seconds=-time.timezone)
        if time.daylight:
            self.DSTOFFSET = timedelta(seconds=-time.altzone)
        else:
            self.DSTOFFSET = self.STDOFFSET

        self.DSTDIFF = self.DSTOFFSET - self.STDOFFSET

    def utcoffset(self, dt):
        if self._isdst(dt):
            return self.DSTOFFSET
        else:
            return self.STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return self.DSTDIFF
        else:
            return self.ZERO

    def tzname(self, dt):
        return time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, 0)
        stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0



#!/usr/bin/env python
"""
_CompleteJob_

Oracle implementation for labeling a job Complete
"""


from WMCore.BossAir.MySQL.CompleteJob import CompleteJob as MySQLCompleteJob

class CompleteJob(MySQLCompleteJob):
    """
    _CompleteJob_

    Label jobs as complete
    """



#!/usr/bin/env python
"""
_SetDatasetAlgo_

Oracle implementation of DBSUpload.SetDatabaseAlgo
Should set the database-algo inDBS switch
"""




from WMComponent.DBS3Buffer.MySQL.SetDatasetAlgo import SetDatasetAlgo as MySQLSetDatasetAlgo

class SetDatasetAlgo(MySQLSetDatasetAlgo):
    """
    Identical to MySQL version

    """



#!/usr/bin/env python
"""
_GetFinishedSubscriptions_

MySQL implementation of Subscription.GetFinishedSubscriptions

Created on Aug 30, 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class GetFinishedSubscriptions(DBFormatter):
    """
    Gets all the subscription marked as finished in the database
    """

    sql = """SELECT id
             FROM wmbs_subscription
             WHERE finished = 1
          """

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        This DAO returns a list of dictionaries containing
        the key 'id' with the id of the finished subscriptions
        """

        result = self.dbi.processData(self.sql,
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_SucceededJobs_

MySQL implementation of Subscriptions.SucceededJobs
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class SucceededJobs(DBFormatter):
    sql = """SELECT id FROM wmbs_job WHERE outcome = 1 and jobgroup IN
             (SELECT id FROM wmbs_jobgroup WHERE subscription = :subscription)"""

    def format(self, result):
        results = DBFormatter.format(self, result)

        subIDs = []
        for row in results:
            subIDs.append(row[0])

        return subIDs

    def execute(self, subscription = None, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"subscription": subscription},
                         conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_GetOpenBlocks_

MySQL implementation of DBS3Buffer.GetOpenBlocks
"""

from WMCore.Database.DBFormatter import DBFormatter

class GetOpenBlocks(DBFormatter):
    sql = """SELECT dbsbuffer_block.blockname AS blockname, dbsbuffer_block.create_time AS create_time,
                    pending_parents.parent_count FROM dbsbuffer_block
               LEFT OUTER JOIN
                 (SELECT dbsbuffer_block.id AS block_id, COUNT(dbsbuffer_block.id) AS parent_count FROM dbsbuffer_block
                    INNER JOIN dbsbuffer_file ON
                      dbsbuffer_block.id = dbsbuffer_file.block_id
                    LEFT OUTER JOIN dbsbuffer_file_parent ON
                      dbsbuffer_file.id = dbsbuffer_file_parent.child
                    LEFT OUTER JOIN dbsbuffer_file dbsbuffer_file_parent_detail ON
                       dbsbuffer_file_parent.parent = dbsbuffer_file_parent_detail.id
                    LEFT OUTER JOIN dbsbuffer_block dbsbuffer_block_parent ON
                      dbsbuffer_file_parent_detail.block_id = dbsbuffer_block_parent.id
                    WHERE dbsbuffer_block_parent.status = 'Open' OR dbsbuffer_block_parent.status = 'Pending'
                    GROUP BY dbsbuffer_block.id) pending_parents ON
                   dbsbuffer_block.id = pending_parents.block_id
             WHERE dbsbuffer_block.status = 'Open' OR dbsbuffer_block.status = 'Pending' AND
                   (pending_parents.parent_count IS NULL or pending_parents.parent_count = 0)"""

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {}, conn = conn,
                                          transaction = transaction)
        return self.formatDict(result)



"""
_Destroy_

Implementation of Destroy for MySQL

"""

from WMCore.Database.DBFormatter import DBFormatter

class Destroy(DBFormatter):

    def execute(self, subscription = None, conn = None, transaction = False):

        sql = """SELECT DATABASE() AS dbname"""

        results = self.dbi.processData(sql, {}, conn = conn,
                                       transaction = transaction)

        dbName = self.formatDict(results)[0]['dbname']

        if dbName == None or dbName == 'None':
            # Then we have no database.
            # This presents us with a problem.  We've been asked to clear a
            # non-existant DB.
            # Obviously we can't drop it, so what we have to do is try
            # to pull the URL from the threaded dbi and use that
            # to create the proper db
            dbName = self.dbi.engine.url.database
        else:
            sql = """DROP DATABASE %s""" % dbName
            self.dbi.processData(sql, {}, conn = conn,
                                 transaction = transaction)

        sql = """CREATE DATABASE %s""" % dbName
        self.dbi.processData(sql, {}, conn = conn,
                             transaction = transaction)

        sql = """USE %s""" % dbName
        self.dbi.processData(sql, {}, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_injectStoreResultsWorkflow_

"""
from __future__ import print_function

import os
import sys
import threading
import pdb
import pprint
import inspect

from WMCore.WMInit import connectToDB
from WMCore.Configuration import loadConfigurationFile

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WorkQueue.WMBSHelper import WMBSHelper
from WMComponent.DBS3Buffer.DBSBufferFile import DBSBufferFile

from WMCore.DataStructs.Run import Run

from WMCore.WMSpec.StdSpecs.StoreResults import storeResultsWorkload, getTestArguments
from DBSAPI.dbsApi import DbsApi

from WMCore.WMSpec.Makers.TaskMaker import TaskMaker

# The default arguments are set in:
#   WMCORE/src/python/WMCore/WMSpec/StdSpecs/StoreResults.py
arguments = getTestArguments()
arguments["UnmergedLFNBase"] = "/store/temp/WMAgent/unmerged"
arguments["MergedLFNBase"] = "/store/temp/results"
arguments["AcquisitionEra"] = 'Blennies'
arguments["CmsPath"] = "/uscmst1/prod/sw/cms"
        #34      arguments.update({
        #35              "CmsPath": "/uscmst1/prod/sw/cms",
        #36              "AcquisitionEra": "Blennies",
arguments["Requestor"] =  "ewv@fnal.gov"
arguments["InputDataset"] = "/MinimumBias/Run2010A-Dec22ReReco_v1/RECO"
arguments["CMSSWVersion"] =  "CMSSW_3_9_7"
arguments["ScramArch"] =  "slc5_ia32_gcc434"
arguments["ProcessingVersion"] =  "v2ewv"
        #42              "CouchDBName": "gutsche_configcache",
arguments["MergedLFNBase"] =  "/store/temp/results"
        #44      })

if len(sys.argv) != 2:
    print("Usage:")
    print("./injectStoreResults.py PROCESSING_VERSION")
    sys.exit(1)
else:
    arguments["ProcessingVersion"] = sys.argv[1]

connectToDB()

workloadName = "StoreResults-%s" % arguments["ProcessingVersion"]
workloadFile = "storeResults-%s.pkl" % arguments["ProcessingVersion"]
os.mkdir(workloadName)
workload = storeResultsWorkload(workloadName, arguments)

# Build a sandbox using TaskMaker
taskMaker = TaskMaker(workload, os.path.join(os.getcwd(), workloadName))
taskMaker.skipSubscription = True
taskMaker.processWorkload()

workload.save(os.path.join(workloadName, workloadFile))


def injectFilesFromDBS(inputFileset, datasetPath):
    """
    _injectFilesFromDBS_

    """
    print("injecting files from %s into %s, please wait..." % (datasetPath, inputFileset.name))
    args={}
    args["url"] = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"
    args["mode"] = "GET"
    dbsApi = DbsApi(args)
    dbsResults = dbsApi.listFileArray(path = datasetPath, retriveList = ["retrive_lumi", "retrive_run"])
    # Limiter on number of files
    dbsResults = dbsResults[0:20]
    print("  found %d files, inserting into wmbs..." % (len(dbsResults)))

    for dbsResult in dbsResults:
        myFile = File(lfn = dbsResult["LogicalFileName"], size = dbsResult["FileSize"],
                      events = dbsResult["NumberOfEvents"], checksums = {"cksum": dbsResult["Checksum"]},
                      locations = "cmssrm.fnal.gov", merged = True)
        myRun = Run(runNumber = dbsResult["LumiList"][0]["RunNumber"])
        for lumi in dbsResult["LumiList"]:
            myRun.appendLumi(lumi["LumiSectionNumber"])
        myFile.addRun(myRun)
        myFile.create()
        inputFileset.addFile(myFile)

        dbsFile = DBSBufferFile(lfn = dbsResult["LogicalFileName"], size = dbsResult["FileSize"],
                                events = dbsResult["NumberOfEvents"], checksums = {"cksum": dbsResult["Checksum"]},
                                locations = "cmssrm.fnal.gov", status = "NOTUPLOADED")
        dbsFile.setDatasetPath(datasetPath)
        dbsFile.setAlgorithm(appName = "cmsRun", appVer = "Unknown", appFam = "Unknown",
                             psetHash = "Unknown", configContent = "Unknown")
        dbsFile.create()

    inputFileset.commit()
    inputFileset.markOpen(False)
    return

myThread = threading.currentThread()
myThread.transaction.begin()
for workloadTask in workload.taskIterator():
    inputFileset = Fileset(name = workloadTask.getPathName())
    inputFileset.create()

    inputDataset = workloadTask.inputDataset()
    inputDatasetPath = "/%s/%s/%s" % (inputDataset.primary,
                                      inputDataset.processed,
                                      inputDataset.tier)
    injectFilesFromDBS(inputFileset, inputDatasetPath)

    myWMBSHelper = WMBSHelper(workload)
    myWMBSHelper._createSubscriptionsInWMBS(workloadTask.getPathName())

myThread.transaction.commit()



#!/usr/bin/env python
"""
_GetValidation_

Oracle implementation of Subscription.GetValidation
"""

from WMCore.WMBS.MySQL.Subscriptions.GetValidation import GetValidation as GetValidationMySQL

class GetValidation(GetValidationMySQL):
    pass
